from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.metagenome.constants import MetagenomeDeliveryType
from cg.services.orders.validation.order_types.metagenome.models.sample import MetagenomeSample


class MetagenomeOrder(OrderWithSamples):
    delivery_type: MetagenomeDeliveryType
    samples: list[MetagenomeSample]

    @property
    def enumerated_samples(self) -> enumerate[MetagenomeSample]:
        return enumerate(self.samples)



from pydantic_core import ErrorDetails, ValidationError

from cg.services.orders.validation.errors.case_errors import CaseError
from cg.services.orders.validation.errors.case_sample_errors import CaseSampleError
from cg.services.orders.validation.errors.order_errors import OrderError
from cg.services.orders.validation.errors.sample_errors import SampleError
from cg.services.orders.validation.errors.validation_errors import ValidationErrors


def convert_errors(pydantic_errors: ValidationError) -> ValidationErrors:
    error_details: list[ErrorDetails] = pydantic_errors.errors()
    order_errors: list[OrderError] = convert_order_errors(error_details)
    case_errors: list[CaseError] = convert_case_errors(error_details=error_details)
    case_sample_errors: list[CaseSampleError] = convert_case_sample_errors(
        error_details=error_details
    )
    sample_errors: list[SampleError] = convert_sample_errors(error_details=error_details)
    return ValidationErrors(
        order_errors=order_errors,
        case_errors=case_errors,
        case_sample_errors=case_sample_errors,
        sample_errors=sample_errors,
    )


def convert_order_errors(error_details: list[ErrorDetails]) -> list[OrderError]:
    errors: list[OrderError] = []
    order_details: list[ErrorDetails] = get_order_error_details(error_details)
    for error_detail in order_details:
        error: OrderError = create_order_error(error_detail)
        errors.append(error)
    return errors


def convert_case_errors(error_details: list[ErrorDetails]) -> list[CaseError]:
    errors: list[CaseError] = []
    case_details: list[ErrorDetails] = get_case_error_details(error_details)
    for error_detail in case_details:
        error: CaseError = create_case_error(error_detail)
        errors.append(error)
    return errors


def convert_sample_errors(error_details: list[ErrorDetails]) -> list[SampleError]:
    errors: list[SampleError] = []
    sample_details: list[ErrorDetails] = get_sample_error_details(error_details)
    for error_detail in sample_details:
        error: SampleError = create_sample_error(error_detail)
        errors.append(error)
    return errors


def create_order_error(error: ErrorDetails) -> OrderError:
    field_name: str = get_order_field_name(error)
    message: str = get_error_message(error)
    error = OrderError(field=field_name, message=message)
    return error


def create_sample_error(error: ErrorDetails) -> SampleError:
    sample_index: int = get_sample_index(error)
    field_name: str = get_sample_field_name(error)
    message: str = get_error_message(error)
    error = SampleError(sample_index=sample_index, field=field_name, message=message)
    return error


def create_case_error(error: ErrorDetails) -> CaseError:
    case_index: int = get_case_index(error=error)
    field_name: str = get_case_field_name(error)
    message: str = get_error_message(error)
    error = CaseError(case_index=case_index, field=field_name, message=message)
    return error


def convert_case_sample_errors(error_details: list[ErrorDetails]) -> list[CaseSampleError]:
    errors: list[CaseSampleError] = []
    case_sample_details: list[ErrorDetails] = get_case_sample_error_details(error_details)
    for error_detail in case_sample_details:
        error = create_case_sample_error(error_detail)
        errors.append(error)
    return errors


def create_case_sample_error(error: ErrorDetails) -> CaseSampleError:
    case_index: int = get_case_index(error=error)
    sample_index: int = get_case_sample_index(error=error)
    field_name: str = get_case_sample_field_name(error)
    message: str = get_error_message(error)
    error = CaseSampleError(
        case_index=case_index,
        sample_index=sample_index,
        field=field_name,
        message=message,
    )
    return error


"""
What follows below are ways of extracting data from a Pydantic ErrorDetails object. The aim is to find out
where the error occurred, for which the 'loc' value (which is a tuple) can be used. It is generally structured in 
alternating strings and ints, specifying field names and list indices. An example:
if loc = ('samples', 2, 'well_position'), that means that the error stems from the well_position of the
third sample in the order.

As an additional point of complexity, the discriminator is also added to the loc, specifically in 
OrdersWithCases which have a discriminator for both cases and samples specifying if it is a new 
or existing case/sample. So
    loc = ('cases', 0, 'new', 'priority')
means that the error concerns the first case in the order, which is a new case, and it concerns the field
'priority'.
"""


def get_sample_error_details(error_details: list[ErrorDetails]) -> list[ErrorDetails]:
    return [error for error in error_details if is_sample_error(error)]


def get_case_error_details(error_details: list[ErrorDetails]) -> list[ErrorDetails]:
    return [error for error in error_details if is_case_error(error)]


def get_case_sample_error_details(error_details: list[ErrorDetails]) -> list[ErrorDetails]:
    return [error for error in error_details if is_case_sample_error(error)]


def get_order_error_details(error_details: list[ErrorDetails]) -> list[ErrorDetails]:
    return [error for error in error_details if is_order_error(error)]


def is_sample_error(error: ErrorDetails) -> bool:
    return error["loc"][0] == "samples"


def is_case_error(error: ErrorDetails) -> bool:
    return "cases" in error["loc"] and "samples" not in error["loc"]


def is_case_sample_error(error: ErrorDetails) -> bool:
    return "cases" in error["loc"] and "samples" in error["loc"]


def is_order_error(error: ErrorDetails) -> bool:
    return error["loc"][0] not in ["cases", "samples"]


def get_error_message(error: ErrorDetails) -> str:
    return error["msg"]


def get_sample_field_name(error: ErrorDetails) -> str:
    index_for_field_name: int = error["loc"].index("samples") + 2
    return error["loc"][index_for_field_name]


def get_case_field_name(error: ErrorDetails) -> str:
    index_for_field_name: int = error["loc"].index("cases") + 3
    return error["loc"][index_for_field_name]


def get_case_sample_field_name(error: ErrorDetails) -> str:
    index_for_field_name: int = error["loc"].index("samples") + 3
    return error["loc"][index_for_field_name]


def get_order_field_name(error: ErrorDetails) -> str:
    return error["loc"][0]


def get_sample_index(error: ErrorDetails) -> int:
    index_for_index: int = error["loc"].index("samples") + 1
    return error["loc"][index_for_index]


def get_case_index(error: ErrorDetails) -> int:
    index_for_index: int = error["loc"].index("cases") + 1
    return error["loc"][index_for_index]


def get_case_sample_index(error: ErrorDetails) -> int:
    index_for_index: int = error["loc"].index("samples") + 1
    return error["loc"][index_for_index]



from pathlib import Path

from cg.models.nf_analysis import WorkflowParameters


class RarediseaseParameters(WorkflowParameters):
    """Model for Raredisease parameters."""

    target_bed_file: str
    analysis_type: str
    save_mapped_as_cram: bool
    vcfanno_extra_resources: str
    vep_filters_scout_fmt: str
    sample_id_map: Path



from abc import ABC, abstractmethod

from cg.services.analysis_starter.configurator.abstract_model import CaseConfig


class Configurator(ABC):

    @abstractmethod
    def configure(self, case_id: str, **flags) -> CaseConfig:
        """Abstract method to configure a case."""
        pass

    @abstractmethod
    def get_config(self, case_id: str, **flags) -> CaseConfig:
        """Abstract method to get the (pre-existing) config for a case."""
        pass



from pathlib import Path

from cg.constants import FileExtensions
from cg.constants.constants import WorkflowManager
from cg.services.analysis_starter.tracker.tracker import Tracker
from cg.store.models import Case


class MicrosaltTracker(Tracker):

    def _workflow_manager(self) -> WorkflowManager:
        return WorkflowManager.Slurm

    def _get_job_ids_path(self, case_id: str) -> Path:
        project_id: str = self._get_lims_project_id(case_id)
        job_ids_path = Path(
            self.workflow_root,
            "results",
            "reports",
            "trailblazer",
            f"{project_id}_slurm_ids{FileExtensions.YAML}",
        )
        # Necessary due to how microsalt structures its output
        self._ensure_old_job_ids_are_removed(job_ids_path)
        return job_ids_path

    def _get_lims_project_id(self, case_id: str) -> str:
        case: Case = self.store.get_case_by_internal_id(case_id)
        sample_id: str = case.links[0].sample.internal_id
        return self._extract_project_id(sample_id)

    @staticmethod
    def _extract_project_id(sample_id: str) -> str:
        return sample_id.rsplit("A", maxsplit=1)[0]

    @staticmethod
    def _ensure_old_job_ids_are_removed(job_ids_path: Path) -> None:
        is_yaml_file: bool = job_ids_path.suffix == FileExtensions.YAML
        if job_ids_path.exists() and is_yaml_file:
            job_ids_path.unlink()



from cg.constants import Workflow
from cg.constants.gene_panel import GenePanelGenomeBuild


def get_genome_build(workflow: Workflow) -> GenePanelGenomeBuild:
    """Return genome build for the given Workflow."""
    workflow_to_genome_build: dict[Workflow, GenePanelGenomeBuild] = {
        Workflow.NALLO: GenePanelGenomeBuild.hg38,
        Workflow.RAREDISEASE: GenePanelGenomeBuild.hg19,
        Workflow.TOMTE: GenePanelGenomeBuild.hg38,
    }
    return workflow_to_genome_build.get(workflow)



"""Code for talking to Scout regarding uploads"""

import logging
from pathlib import Path
from subprocess import CalledProcessError

from cg.apps.scout.scout_export import ScoutExportCase, Variant
from cg.constants.constants import FileFormat
from cg.constants.gene_panel import GENOME_BUILD_37
from cg.constants.scout import ScoutCustomCaseReportTags
from cg.exc import ScoutUploadError
from cg.io.controller import ReadFile, ReadStream
from cg.models.scout.scout_load_config import ScoutLoadConfig
from cg.services.slurm_upload_service.slurm_upload_service import SlurmUploadService
from cg.utils.commands import Process

LOG = logging.getLogger(__name__)


class ScoutAPI:
    """Interface to Scout."""

    def __init__(self, scout_config, slurm_upload_service: SlurmUploadService):
        binary_path = scout_config.binary_path
        config_path = scout_config.config_path
        self.process = Process(binary=binary_path, config=config_path)
        self.slurm_upload_service = slurm_upload_service
        self.scout_base_command = f"{binary_path} --config {config_path}"

    def upload(self, scout_load_config: Path, force: bool = False) -> None:
        """Load analysis of a new family into Scout."""

        scout_config: dict = ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=scout_load_config
        )
        scout_load_config_object = ScoutLoadConfig(**scout_config)
        case_id: str = scout_load_config_object.family
        existing_case: ScoutExportCase | None = self.get_case(case_id)
        load_command = ["load", "case", str(scout_load_config)]
        if existing_case:
            if force or scout_load_config_object.analysis_date > existing_case.analysis_date:
                load_command.append("--update")
                LOG.info("update existing Scout case")
            else:
                existing_date = existing_case.analysis_date.date()
                LOG.warning(f"Analysis of case already loaded: {existing_date}")
                return
        LOG.debug("load new Scout case")
        self.process.run_command(load_command)
        LOG.debug("Case loaded successfully to Scout")

    def export_panels(
        self, panels: list[str], build: str = GENOME_BUILD_37, dry_run: bool = False
    ) -> list[str]:
        """Pass through to export of a list of gene panels.

        Return list of lines in bed format
        """
        export_panels_command = ["export", "panel", "--bed"]
        export_panels_command.extend(iter(panels))

        if build:
            export_panels_command.extend(["--build", build])

        try:
            self.process.run_command(export_panels_command, dry_run=dry_run)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.info("Could not find panels")
            return []

        return list(self.process.stdout_lines())

    def export_managed_variants(self, genome_build: str = GENOME_BUILD_37) -> list[str]:
        """Export a list of managed variants."""
        export_command = ["export", "managed"]
        if genome_build:
            export_command.extend(["--build", genome_build])
        try:
            self.process.run_command(export_command)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.info("Could not export managed variants")
            return []
        return list(self.process.stdout_lines())

    def get_genes(self, panel_id: str, build: str = None) -> list[dict]:
        """Return panel genes."""
        export_panel_command = ["export", "panel", panel_id]
        if build:
            export_panel_command.extend(["--build", build])

        try:
            self.process.run_command(export_panel_command)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.info(f"Could not find panel {panel_id}")
            return []

        panel_genes = []
        for gene_line in self.process.stdout_lines():
            if gene_line.startswith("#"):
                continue
            gene_info = gene_line.strip().split("\t")
            if len(gene_info) <= 1:
                continue
            panel_genes.append({"hgnc_id": int(gene_info[0]), "hgnc_symbol": gene_info[1]})

        return panel_genes

    def get_causative_variants(self, case_id: str) -> list[Variant]:
        """
        Get causative variants for a case.
        """
        get_causatives_command = ["export", "variants", "--json", "--case-id", case_id]
        try:
            self.process.run_command(get_causatives_command)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.warning(f"Could not find case {case_id} in scout")
            return []
        variants: list[Variant] = [
            Variant(**variant_info)
            for variant_info in ReadStream.get_content_from_stream(
                file_format=FileFormat.JSON, stream=self.process.stdout
            )
        ]

        return variants

    def get_case(self, case_id: str) -> ScoutExportCase | None:
        """Fetch a case from Scout"""
        cases: list[ScoutExportCase] = self.get_cases(case_id=case_id)
        return cases[0] if cases else None

    def get_cases(
        self,
        case_id: str | None = None,
        reruns: bool = False,
        finished: bool = False,
        status: str | None = None,
        days_ago: int = None,
    ) -> list[ScoutExportCase]:
        """Interact with cases existing in the database."""
        get_cases_command = ["export", "cases", "--json"]
        if case_id:
            get_cases_command.extend(["--case-id", case_id])

        elif status:
            get_cases_command.extend(["--status", status])

        elif finished:
            get_cases_command.append("--finished")

        if reruns:
            LOG.debug("Fetching cases that are reruns")
            get_cases_command.append("--reruns")

        if days_ago:
            get_cases_command.extend(["--within-days", str(days_ago)])

        try:
            self.process.run_command(get_cases_command)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.info("Could not find cases")
            return []

        cases: list[ScoutExportCase] = []
        for case_export in ReadStream.get_content_from_stream(
            file_format=FileFormat.JSON, stream=self.process.stdout
        ):
            LOG.debug(f"Validating case {case_export.get('_id')}")
            cases.append(ScoutExportCase.model_validate(case_export))
        return cases

    def get_solved_cases(self, days_ago: int) -> list[ScoutExportCase]:
        """
        Get cases solved within chosen time span.
        """
        return self.get_cases(status="solved", days_ago=days_ago)

    def upload_delivery_report(self, report_path: str, case_id: str, update: bool = False) -> None:
        """Load a delivery report into a case in the database.
        If the report already exists the function will exit.
        If the user want to load a report that is already in the database
        'update' has to be 'True'."""

        upload_command: list[str] = ["load", "delivery-report", case_id, report_path]

        if update:
            upload_command.append("--update")

        try:
            LOG.info(f"Uploading delivery report {report_path} to case {case_id}")
            self.process.run_command(upload_command)
        except CalledProcessError:
            LOG.warning("Something went wrong when uploading delivery report")

    def upload_report(self, case_id: str, report_path: str, report_type: str) -> None:
        """Load report into a case in the database."""

        upload_report_command: list[str] = [
            "load",
            "report",
            "-t",
            report_type,
            case_id,
            report_path,
        ]

        try:
            LOG.info(f"Uploading {report_type} report to case {case_id}")
            self.process.run_command(upload_report_command)
        except CalledProcessError:
            LOG.warning(f"Something went wrong when uploading {report_type} for case {case_id}")

    def upload_fusion_report(self, case_id: str, report_path: str, research: bool) -> None:
        """Load a fusion report into a case in the database."""

        report_type: str = (
            ScoutCustomCaseReportTags.GENE_FUSION_RESEARCH
            if research
            else ScoutCustomCaseReportTags.GENE_FUSION
        )
        self.upload_report(case_id=case_id, report_path=report_path, report_type=report_type)

    def upload_splice_junctions_bed(
        self, file_path: str, case_id: str, customer_sample_id: str
    ) -> None:
        """Load a splice junctions bed file into a case in the database."""

        upload_command: list[str] = [
            "update",
            "individual",
            "--case-id",
            case_id,
            "--ind",
            customer_sample_id,
            "splice_junctions_bed",
            file_path,
        ]

        try:
            LOG.info(f"Uploading splice junctions bed file {file_path} to case {case_id}.")
            self.process.run_command(upload_command)
        except CalledProcessError as error:
            raise ScoutUploadError(
                "Something went wrong when uploading the splice junctions bed file."
            ) from error

    def upload_rna_coverage_bigwig(
        self, file_path: str, case_id: str, customer_sample_id: str
    ) -> None:
        """Load a rna coverage bigwig file into a case in the database."""

        upload_command: list[str] = [
            "update",
            "individual",
            "--case-id",
            case_id,
            "--ind",
            customer_sample_id,
            "rna_coverage_bigwig",
            file_path,
        ]

        try:
            LOG.info(f"Uploading rna coverage bigwig file {file_path} to case {case_id}")
            self.process.run_command(upload_command)
        except CalledProcessError as error:
            raise ScoutUploadError(
                "Something went wrong when uploading rna coverage bigwig file"
            ) from error

    def upload_omics_sample_id(
        self, dna_case_id: str, customer_sample_id: str, rna_sample_internal_id: str
    ) -> None:
        """Load a omics sample id into a case in the database."""

        upload_command: list[str] = [
            "update",
            "individual",
            "--case-id",
            dna_case_id,
            "--ind",
            customer_sample_id,
            "omics_sample_id",
            rna_sample_internal_id,
        ]
        try:
            LOG.info("Uploading omics_sample_id")
            self.process.run_command(upload_command)
        except CalledProcessError as error:
            raise ScoutUploadError("Something went wrong when uploading omics_sample_id") from error

    def upload_rna_fraser_outrider(
        self,
        case_id: str,
        fraser_file_path: str | None,
        outrider_file_path: str | None,
        customer_case_name: str,
        cust_id: str,
    ) -> None:
        """Load a rna fraser file into a case in the database."""
        upload_command: list[str] = [
            "update",
            "case",
            "-n",
            customer_case_name,
            "-i",
            cust_id,
        ]
        if fraser_file_path:
            upload_command.extend(["--fraser", fraser_file_path])
            LOG.info(f"Uploading rna fraser file {fraser_file_path} to case {case_id}")

        if outrider_file_path:
            upload_command.extend(["--outrider", outrider_file_path])
            LOG.info(f"Uploading outrider file {outrider_file_path} to case {case_id}")
        try:
            self.process.run_command(upload_command)
        except CalledProcessError as error:
            raise ScoutUploadError("Something went wrong when uploading rna fraser file") from error

    def upload_rna_genome_build(
        self, case_id: str, customer_case_name: str, cust_id: str, rna_genome_build: str
    ) -> None:
        """Load a rna fraser file into a case in the database."""

        upload_command: list[str] = [
            "update",
            "case",
            "-n",
            customer_case_name,
            "-i",
            cust_id,
            "--rna-genome-build",
            rna_genome_build,
        ]
        try:
            LOG.info(f"Updating rna genome build {rna_genome_build} to case {case_id}")
            self.process.run_command(upload_command)
        except CalledProcessError as error:
            raise ScoutUploadError("Something went wrong when updating rna genome build") from error

    def upload_rna_delivery_report(self, case_id: str, report_path: str) -> None:
        """Load a RNA report into a case in the database."""

        self.upload_report(
            case_id=case_id,
            report_path=report_path,
            report_type=ScoutCustomCaseReportTags.RNA_DELIVERY,
        )

    def load_variant_outlier(self, case_id: str) -> None:
        """Load a rna fraser file into a case in the database."""

        upload_command: list[str] = [
            "load",
            "variants",
            "--outlier",
            case_id,
        ]
        try:
            LOG.info(f"Loading variants outlier for case {case_id}")
            self.process.run_command(upload_command)
        except CalledProcessError as error:
            raise ScoutUploadError("Something went wrong when loading variants outlier") from error

    def upload_rna_alignment_file(
        self, case_id: str, customer_sample_id: str, file_path: str
    ) -> None:
        """Load an RNA alignment CRAM file into a case in the database."""

        upload_command: list[str] = [
            "update",
            "individual",
            "--case-id",
            case_id,
            "--ind",
            customer_sample_id,
            "rna_alignment_path",
            file_path,
        ]

        try:
            LOG.info(f"Uploading RNA alignment CRAM file {file_path} to case {case_id}")
            self.process.run_command(upload_command)
        except CalledProcessError as error:
            raise ScoutUploadError(
                "Something went wrong when uploading the RNA alignment CRAM file"
            ) from error



from pathlib import Path

from cg.constants.constants import FileExtensions, WorkflowManager
from cg.services.analysis_starter.tracker.tracker import Tracker


class NextflowTracker(Tracker):
    def _workflow_manager(self) -> WorkflowManager:
        return WorkflowManager.Tower

    def _get_job_ids_path(self, case_id: str) -> Path:
        """Return the path to a Trailblazer config file containing Tower IDs."""
        return Path(self.workflow_root, case_id, "tower_ids").with_suffix(FileExtensions.YAML)



from pathlib import Path

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.exc import CaseNotConfiguredError
from cg.models.cg_config import CommonAppConfig
from cg.services.analysis_starter.configurator.configurator import Configurator
from cg.services.analysis_starter.configurator.extensions.abstract import PipelineExtension
from cg.services.analysis_starter.configurator.file_creators.nextflow.config_file import (
    NextflowConfigFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.abstract import (
    ParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.abstract import (
    NextflowSampleSheetCreator,
)
from cg.services.analysis_starter.configurator.models.nextflow import NextflowCaseConfig
from cg.store.store import Store


class NextflowConfigurator(Configurator):
    def __init__(
        self,
        config_file_creator: NextflowConfigFileCreator,
        housekeeper_api: HousekeeperAPI,
        lims: LimsAPI,
        params_file_creator: ParamsFileCreator,
        pipeline_config: CommonAppConfig,
        sample_sheet_creator: NextflowSampleSheetCreator,
        store: Store,
        pipeline_extension: PipelineExtension = PipelineExtension(),
    ):
        self.root_dir: str = pipeline_config.root
        self.pipeline_repository = pipeline_config.repository
        self.pipeline_revision = pipeline_config.revision
        self.config_profiles = [pipeline_config.profile]
        self.pre_run_script = pipeline_config.pre_run_script
        self.store: Store = store
        self.housekeeper_api: HousekeeperAPI = housekeeper_api
        self.lims: LimsAPI = lims
        self.config_file_creator = config_file_creator
        self.pipeline_extension = pipeline_extension
        self.sample_sheet_creator = sample_sheet_creator
        self.params_file_creator = params_file_creator

    def configure(self, case_id: str, **flags) -> NextflowCaseConfig:
        """Configure a Nextflow case so that it is ready for analysis. This entails
        1. Creating a case directory.
        2. Creating a sample sheet.
        3. Creating a parameters file.
        4. Creating a configuration file.
        5. Creating any pipeline specific files."""
        case_path: Path = self._get_case_path(case_id)
        self._create_case_directory(case_id)
        self.sample_sheet_creator.create(case_id=case_id, case_path=case_path)
        sample_sheet_path: Path = self.sample_sheet_creator.get_file_path(
            case_id=case_id, case_path=case_path
        )
        self.params_file_creator.create(
            case_id=case_id, case_path=case_path, sample_sheet_path=sample_sheet_path
        )
        self.config_file_creator.create(case_id=case_id, case_path=case_path)
        self.pipeline_extension.configure(case_id=case_id, case_path=case_path)
        return self.get_config(case_id=case_id, **flags)

    def get_config(self, case_id: str, **flags) -> NextflowCaseConfig:
        """
        Gets the configuration properties for the provided case.
        Raises:
            CaseNotConfiguredError if the params file or config file does not exist.
        """
        case_path: Path = self._get_case_path(case_id=case_id)
        params_file_path: Path = self.params_file_creator.get_file_path(
            case_id=case_id, case_path=case_path
        )
        config_file_path: Path = self.config_file_creator.get_file_path(
            case_id=case_id, case_path=case_path
        )
        if not params_file_path.exists() or not config_file_path.exists():
            raise CaseNotConfiguredError(
                f"Please ensure that both the parameters file {params_file_path.as_posix()} and the configuration file {config_file_path.as_posix()} exists."
            )
        config = NextflowCaseConfig(
            case_id=case_id,
            case_priority=self.store.get_case_priority(case_id),
            config_profiles=self.config_profiles,
            nextflow_config_file=config_file_path.as_posix(),
            params_file=params_file_path.as_posix(),
            pipeline_repository=self.pipeline_repository,
            pre_run_script=self.pre_run_script,
            revision=self.pipeline_revision,
            stub_run=False,
            work_dir=self._get_work_dir(case_id).as_posix(),
            workflow=self.store.get_case_workflow(case_id),
        )
        return config.model_copy(update=flags)

    def _get_case_path(self, case_id: str) -> Path:
        """Path to case working directory."""
        return Path(self.root_dir, case_id)

    def _create_case_directory(self, case_id: str) -> None:
        """Create case working directory."""
        case_path: Path = self._get_case_path(case_id=case_id)
        case_path.mkdir(parents=True, exist_ok=True)

    def _get_work_dir(self, case_id: str) -> Path:
        return Path(self.root_dir, case_id, "work")



import logging

from pydantic import BeforeValidator, Field, PrivateAttr, model_validator
from typing_extensions import Annotated

from cg.models.orders.sample_base import ContainerEnum, ControlEnum, PriorityEnum
from cg.services.orders.validation.constants import IndexEnum
from cg.services.orders.validation.index_sequences import INDEX_SEQUENCES
from cg.services.orders.validation.models.sample import Sample
from cg.services.orders.validation.utils import parse_control

LOG = logging.getLogger(__name__)


class RMLSample(Sample):
    concentration: float
    concentration_sample: float | None = None
    container: ContainerEnum | None = Field(default=None, exclude=True)
    control: Annotated[ControlEnum, BeforeValidator(parse_control)] = ControlEnum.not_control
    index: IndexEnum
    index_number: int | None = None
    _index_sequence: str | None = PrivateAttr(default=None)
    pool: str
    priority: PriorityEnum
    rml_plate_name: str | None = None
    volume: int
    well_position_rml: str | None = None

    @model_validator(mode="after")
    def set_default_index_sequence(self) -> "RMLSample":
        """Set a default index_sequence from the index and index_number."""
        if self.index and self.index_number:
            try:
                self._index_sequence = INDEX_SEQUENCES[self.index][self.index_number - 1]
            except Exception:
                LOG.warning(
                    f"No index sequence set and no suitable sequence found for index {self.index}, number {self.index_number}"
                )
        return self

    def model_dump(self, **kwargs) -> dict:
        data = super().model_dump(**kwargs)
        data["index_sequence"] = self._index_sequence
        return data



import logging

import requests

from cg.constants.priority import SlurmQos
from cg.models.cg_config import SeqeraPlatformConfig
from cg.services.analysis_starter.submitters.seqera_platform.dtos import WorkflowLaunchRequest

LOG = logging.getLogger(__name__)


class SeqeraPlatformClient:
    def __init__(self, config: SeqeraPlatformConfig):
        self.base_url: str = config.base_url
        self.bearer_token: str = config.bearer_token
        self.auth_headers: dict = {"Authorization": f"Bearer {self.bearer_token}"}
        self.compute_environment_ids: dict[SlurmQos, str] = config.compute_environments
        self.workspace_id: int = config.workspace_id

    def run_case(self, request: WorkflowLaunchRequest) -> str:
        """Launches a case from the request and returns the workflow ID."""
        url = f"{self.base_url}/workflow/launch"
        params: dict = {"workspaceId": self.workspace_id}
        LOG.debug(
            f"Sending request body {request.model_dump()} \n Headers: {self.auth_headers} \n Params: {params}"
        )
        response: requests.Response = requests.post(
            url=url,
            headers=self.auth_headers,
            params=params,
            json=request.model_dump(),
        )
        response.raise_for_status()
        return response.json()["workflowId"]



"""Post-processing service abstract classes."""

import logging
from abc import ABC, abstractmethod
from pathlib import Path

from cg.services.run_devices.abstract_models import PostProcessingDTOs, RunData, RunMetrics
from cg.services.run_devices.constants import POST_PROCESSING_COMPLETED

LOG = logging.getLogger(__name__)


class RunDataGenerator(ABC):
    """Abstract class that holds functionality to create a run data object."""

    @abstractmethod
    def get_run_data(self, run_name: str, sequencing_dir: str) -> RunData:
        """Get the run data for a sequencing run."""
        pass


class RunValidator(ABC):
    """Abstract class that holds functionality to validate a run and ensure it can start post processing."""

    @abstractmethod
    def ensure_post_processing_can_start(self, run_data: RunData):
        """Ensure a post processing run can start."""
        pass

    @abstractmethod
    def validate_run_files(self, run_data: RunData):
        """Validate presence run files."""
        pass


class RunFileManager(ABC):
    """Abstract class that manages files related to an instrument run."""

    @abstractmethod
    def get_files_to_parse(self, run_data: RunData) -> list[Path]:
        """Get the files required for the PostProcessingMetricsService."""
        pass

    @abstractmethod
    def get_files_to_store(self, run_data: RunData) -> list[Path]:
        """Get the files to store for the PostProcessingHKService."""
        pass


class PostProcessingMetricsParser(ABC):
    """Abstract class that manages the metrics parsing related to an instrument run."""

    @abstractmethod
    def parse_metrics(self, run_data: RunData) -> RunMetrics:
        """Parse the sequencing run metrics."""
        pass


class PostProcessingDataTransferService(ABC):
    """Abstract class that manages data transfer from parsed metrics to the database structure."""

    @abstractmethod
    def get_post_processing_dtos(self, run_data: RunData) -> PostProcessingDTOs:
        """Get the data transfer objects for the PostProcessingStoreService."""
        pass


class PostProcessingStoreService(ABC):
    """Abstract class that manages storing data transfer objects in the database."""

    @abstractmethod
    def store_post_processing_data(self, run_data: RunData, dry_run: bool = False):
        """Store the data transfer objects in the database."""
        pass


class PostProcessingHKService(ABC):
    """Abstract class that manages storing of files for an instrument run."""

    @abstractmethod
    def store_files_in_housekeeper(self, run_data: RunData, dry_run: bool = False):
        """Store the files in Housekeeper."""
        pass


class PostProcessingService(ABC):
    """Abstract class that encapsulates the logic required for post-processing a sequencing run."""

    @abstractmethod
    def post_process(self, run_name: str, dry_run: bool = False):
        """Store sequencing metrics in StatusDB and relevant files in Housekeeper."""
        pass

    @abstractmethod
    def is_run_processed(self, run_name: str) -> bool:
        """Check if a run has been post-processed."""
        pass

    @abstractmethod
    def can_post_processing_start(self, run_name: str) -> bool:
        """Check if a run can be post processed."""
        pass

    @staticmethod
    def _touch_post_processing_complete(run_data: RunData, dry_run: bool = False) -> None:
        """Touch the post-processing complete file."""
        processing_complete_file = Path(run_data.full_path, POST_PROCESSING_COMPLETED)
        if not dry_run:
            processing_complete_file.touch()
        LOG.debug(f"Post-processing complete for {run_data.full_path}")



from cg.constants.priority import SlurmQos
from cg.services.analysis_starter.configurator.abstract_model import CaseConfig


class NextflowCaseConfig(CaseConfig):
    case_priority: SlurmQos
    config_profiles: list[str]
    nextflow_config_file: str
    params_file: str
    pipeline_repository: str
    pre_run_script: str
    revision: str
    stub_run: bool
    work_dir: str



from enum import StrEnum
from pathlib import Path

from pydantic import BaseModel, field_validator

from cg.constants import SexOptions
from cg.exc import NfSampleSheetError
from cg.models.nf_analysis import WorkflowParameters
from cg.models.qc_metrics import QCMetrics


class NalloQCMetrics(QCMetrics):
    """Nallo QC metrics."""

    avg_sequence_length: float | None
    coverage_bases: float | None
    median_coverage: float | None
    percent_duplicates: float | None
    predicted_sex_sex_check: SexOptions


class NalloSampleSheetEntry(BaseModel):
    """Nallo sample model is used when building the sample sheet."""

    project: str
    sample: str
    read_file: Path
    family_id: str
    paternal_id: str
    maternal_id: str
    sex: int
    phenotype: int

    @property
    def reformat_sample_content(self) -> list[list[str]]:
        """Reformat sample sheet content as a list of lists, where each list represents a line in the final file."""
        return [
            [
                self.project,
                self.sample,
                self.read_file,
                self.family_id,
                self.paternal_id,
                self.maternal_id,
                self.sex,
                self.phenotype,
            ]
        ]

    @field_validator("read_file")
    @classmethod
    def read_file_exists(cls, bam_path: Path) -> Path:
        """Verify that bam files exist."""
        if not bam_path.is_file():
            raise NfSampleSheetError(f"Bam file does not exist: {str(bam_path)}")
        return bam_path


class NalloSampleSheetHeaders(StrEnum):
    project: str = "project"
    sample: str = "sample"
    file: str = "file"
    family_id: str = "family_id"
    paternal_id: str = "paternal_id"
    maternal_id: str = "maternal_id"
    sex: str = "sex"
    phenotype: str = "phenotype"

    @classmethod
    def list(cls) -> list[str]:
        return list(map(lambda header: header.value, cls))


class NalloParameters(WorkflowParameters):
    """Model for Nallo parameters."""

    filter_variants_hgnc_ids: str



from pydantic import BeforeValidator, Field, PrivateAttr
from typing_extensions import Annotated

from cg.models.orders.sample_base import ControlEnum, PriorityEnum
from cg.services.orders.validation.constants import ElutionBuffer, ExtractionMethod
from cg.services.orders.validation.models.sample import Sample
from cg.services.orders.validation.utils import parse_buffer, parse_control, parse_extraction_method


class MicrosaltSample(Sample):
    control: Annotated[ControlEnum, BeforeValidator(parse_control)] = ControlEnum.not_control
    elution_buffer: Annotated[ElutionBuffer, BeforeValidator(parse_buffer)]
    extraction_method: Annotated[ExtractionMethod, BeforeValidator(parse_extraction_method)]
    organism: str
    organism_other: str | None = None
    priority: PriorityEnum
    reference_genome: str = Field(max_length=255)
    _verified_organism: bool | None = PrivateAttr(default=None)

    def model_dump(self, **kwargs) -> dict:
        data = super().model_dump(**kwargs)
        data["verified_organism"] = self._verified_organism
        return data



"""Abstract base classes for order submitters."""

import logging
from abc import ABC, abstractmethod

from cg.services.orders.lims_service.service import OrderLimsService
from cg.services.orders.validation.models.order import Order
from cg.services.orders.validation.models.sample import Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class StoreOrderService(ABC):
    @abstractmethod
    def __init__(self, status_db: Store, lims_service: OrderLimsService):
        self.status_db = status_db
        self.lims = lims_service

    @abstractmethod
    def store_order(self, order: Order):
        pass

    @staticmethod
    def _fill_in_sample_ids(samples: list[Sample], lims_map: dict) -> None:
        """Fill in LIMS sample ids."""
        for sample in samples:
            LOG.debug(f"{sample.name}: link sample to LIMS")
            internal_id = lims_map[sample.name]
            LOG.info(f"{sample.name} -> {internal_id}: connect sample to LIMS")
            sample._generated_lims_id = internal_id



from cg.services.orders.validation.constants import MAXIMUM_VOLUME, MINIMUM_VOLUME, IndexEnum
from cg.services.orders.validation.errors.order_errors import OrderError
from cg.services.orders.validation.index_sequences import INDEX_SEQUENCES


class SampleError(OrderError):
    sample_index: int


class ApplicationNotValidError(SampleError):
    field: str = "application"
    message: str = "Chosen application does not exist"


class ApplicationArchivedError(SampleError):
    field: str = "application"
    message: str = "Chosen application is archived"


class ApplicationNotCompatibleError(SampleError):
    field: str = "application"
    message: str = "Chosen application is not compatible with workflow"


class OccupiedWellError(SampleError):
    field: str = "well_position"
    message: str = "Well is already occupied"


class WellPositionMissingError(SampleError):
    field: str = "well_position"
    message: str = "Well position is required for well plates"


class WellPositionRmlMissingError(SampleError):
    field: str = "well_position_rml"
    message: str = "Well position is required for RML plates"


class SampleNameRepeatedError(SampleError):
    field: str = "name"
    message: str = "Sample name repeated"


class InvalidVolumeError(SampleError):
    field: str = "volume"
    message: str = f"Volume must be between {MINIMUM_VOLUME}-{MAXIMUM_VOLUME} μL"


class VolumeRequiredError(SampleError):
    field: str = "volume"
    message: str = "Volume is required"


class SampleNameNotAvailableError(SampleError):
    field: str = "name"
    message: str = "Sample name already used in previous order"


class SampleNameNotAvailableControlError(SampleError):
    field: str = "name"
    message: str = "Sample name already in use. Only control samples are allowed repeated names"


class ContainerNameRepeatedError(SampleError):
    field: str = "container_name"
    message: str = "Tube names must be unique among samples"


class WellFormatError(SampleError):
    field: str = "well_position"
    message: str = "Well position must follow the format A-H:1-12"


class WellFormatRmlError(SampleError):
    field: str = "well_position_rml"
    message: str = "Well position must follow the format A-H:1-12"


class ContainerNameMissingError(SampleError):
    field: str = "container_name"
    message: str = "Container must have a name"


class BufferInvalidError(SampleError):
    field: str = "elution_buffer"
    message: str = "Buffer must be Tris-HCl or Nuclease-free water when skipping reception control."


class ConcentrationRequiredError(SampleError):
    field: str = "concentration_ng_ul"
    message: str = "Concentration is required when skipping reception control."


class ConcentrationInvalidIfSkipRCError(SampleError):
    def __init__(self, sample_index: int, allowed_interval: tuple[float, float]):
        field: str = "concentration_ng_ul"
        message: str = (
            f"Concentration must be between {allowed_interval[0]} ng/μL and "
            f"{allowed_interval[1]} ng/μL if reception control should be skipped"
        )
        super(SampleError, self).__init__(sample_index=sample_index, field=field, message=message)


class PoolApplicationError(SampleError):
    def __init__(self, sample_index: int, pool_name: str):
        field: str = "application"
        message: str = f"Multiple applications detected in pool {pool_name}"
        super(SampleError, self).__init__(sample_index=sample_index, field=field, message=message)


class PoolPriorityError(SampleError):
    def __init__(self, sample_index: int, pool_name: str):
        field: str = "priority"
        message: str = f"Multiple priorities detected in pool {pool_name}"
        super(SampleError, self).__init__(sample_index=sample_index, field=field, message=message)


class IndexNumberMissingError(SampleError):
    field: str = "index_number"
    message: str = "Index number is required"


class IndexNumberOutOfRangeError(SampleError):
    def __init__(self, sample_index: int, index: IndexEnum):
        field: str = "index_number"
        maximum: int = len(INDEX_SEQUENCES[index])
        message: str = f"Index number must be a number between 1 and {maximum}"
        super(SampleError, self).__init__(sample_index=sample_index, field=field, message=message)


class CaptureKitMissingError(SampleError):
    field: str = "capture_kit"
    message: str = "Bait set is required for TGS analyses"


class CaptureKitInvalidError(SampleError):
    field: str = "capture_kit"
    message: str = "Bait set must be valid"



from pathlib import Path

from cg.apps.lims import LimsAPI
from cg.constants import FileExtensions
from cg.constants.scout import ScoutExportFileName
from cg.constants.symbols import EMPTY_STRING
from cg.exc import CgDataError
from cg.io.csv import write_csv
from cg.io.yaml import read_yaml, write_yaml_nextflow_style
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.abstract import (
    ParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.models import (
    RarediseaseParameters,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.utils import (
    replace_values_in_params_file,
)
from cg.store.models import BedVersion, Case, Sample
from cg.store.store import Store


class RarediseaseParamsFileCreator(ParamsFileCreator):

    def __init__(self, store: Store, lims: LimsAPI, params: str):
        self.store = store
        self.lims = lims
        self.params = params

    @staticmethod
    def get_file_path(case_id: str, case_path: Path) -> Path:
        """Return the path to the params file."""
        return Path(case_path, f"{case_id}_params_file").with_suffix(FileExtensions.YAML)

    def create(self, case_id: str, case_path: Path, sample_sheet_path: Path) -> None:
        """Create the params file for a case."""
        file_path: Path = self.get_file_path(case_id=case_id, case_path=case_path)
        content: dict = self._get_content(
            case_id=case_id, case_path=case_path, sample_sheet_path=sample_sheet_path
        )
        write_yaml_nextflow_style(file_path=file_path, content=content)

    def _get_content(self, case_id: str, case_path: Path, sample_sheet_path: Path) -> dict:
        """Return the content of a params file for a case."""
        case_workflow_parameters: dict = self._get_case_parameters(
            case_id=case_id, case_path=case_path, sample_sheet_path=sample_sheet_path
        ).model_dump()
        workflow_parameters: dict = read_yaml(Path(self.params))
        duplicate_keys = set(case_workflow_parameters.keys()) & set(workflow_parameters.keys())
        if duplicate_keys:
            raise CgDataError(f"Duplicate parameter keys found: {duplicate_keys}")
        parameters: dict = case_workflow_parameters | workflow_parameters
        curated_parameters: dict = replace_values_in_params_file(parameters)
        return curated_parameters

    def _get_case_parameters(
        self, case_id: str, case_path: Path, sample_sheet_path: Path
    ) -> RarediseaseParameters:
        """Return case-specific parameters for the analysis."""
        analysis_type: str = self._get_data_analysis_type(case_id)
        target_bed_file: str = self._get_target_bed_from_lims(case_id)
        sample_mapping_file: Path = self._create_sample_mapping_file(
            case_id=case_id, case_path=case_path
        )
        return RarediseaseParameters(
            input=sample_sheet_path,
            outdir=case_path,
            analysis_type=analysis_type,
            target_bed_file=target_bed_file,
            save_mapped_as_cram=True,
            vcfanno_extra_resources=f"{case_path}/{ScoutExportFileName.MANAGED_VARIANTS}",
            vep_filters_scout_fmt=f"{case_path}/{ScoutExportFileName.PANELS}",
            sample_id_map=sample_mapping_file,
        )

    def _get_data_analysis_type(self, case_id: str) -> str:
        """
        Return case analysis type (WEG, WGS, WTS or TGS). Assumes all case samples have the same
        analysis type.
        """
        sample: Sample = self.store.get_samples_by_case_id(case_id=case_id)[0]
        return sample.application_version.application.analysis_type

    def _get_target_bed_from_lims(self, case_id: str) -> str:
        """
        Get target bed filename from LIMS. Return an empty string if no target bed is found in LIMS.
        Raises:
            CgDataError: if the bed target capture version is not found in StatusDB.
        """
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        sample: Sample = case.samples[0]
        if sample.from_sample:
            sample: Sample = self.store.get_sample_by_internal_id(internal_id=sample.from_sample)
        target_bed_shortname: str | None = self.lims.capture_kit(lims_id=sample.internal_id)
        if not target_bed_shortname:
            return EMPTY_STRING
        bed_version: BedVersion | None = self.store.get_bed_version_by_short_name(
            bed_version_short_name=target_bed_shortname
        )
        if not bed_version:
            raise CgDataError(f"Bed-version {target_bed_shortname} does not exist")
        return bed_version.filename

    def _create_sample_mapping_file(self, case_id: str, case_path: Path) -> Path:
        """Create a sample mapping file for the case and returns its path."""
        file_path = Path(case_path, f"{case_id}_customer_internal_mapping").with_suffix(
            FileExtensions.CSV
        )
        content: list[list[str]] = [["customer_id", "internal_id"]]
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        for sample in case.samples:
            content.append([sample.name, sample.internal_id])
        write_csv(file_path=file_path, content=content)
        return file_path



"""CLI support to create config and/or start NALLO."""

import logging

import rich_click as click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS, echo_lines
from cg.cli.workflow.commands import ARGUMENT_CASE_ID
from cg.constants.cli_options import DRY_RUN
from cg.cli.workflow.nf_analysis import (
    config_case,
    metrics_deliver,
    report_deliver,
    run,
    start,
    start_available,
    store,
    store_available,
    store_housekeeper,
)

from cg.constants.constants import MetaApis
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.nallo import NalloAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True, context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_context
def nallo(context: click.Context) -> None:
    """GMS/Nallo analysis workflow."""
    AnalysisAPI.get_help(context)
    context.obj.meta_apis[MetaApis.ANALYSIS_API] = NalloAnalysisAPI(config=context.obj)


nallo.add_command(config_case)
nallo.add_command(report_deliver)
nallo.add_command(run)
nallo.add_command(start)
nallo.add_command(start_available)
nallo.add_command(store)
nallo.add_command(store_available)
nallo.add_command(store_housekeeper)
nallo.add_command(metrics_deliver)


@nallo.command("panel")
@DRY_RUN
@ARGUMENT_CASE_ID
@click.pass_obj
def panel(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Write aggregated gene panel file exported from Scout."""

    analysis_api: NalloAnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.status_db.verify_case_exists(case_internal_id=case_id)

    bed_lines: list[str] = analysis_api.get_gene_panel(case_id=case_id)
    if dry_run:
        echo_lines(lines=bed_lines)
        return
    analysis_api.write_panel_as_tsv(case_id=case_id, content=bed_lines)



from pathlib import Path

from cg.services.analysis_starter.configurator.extensions.abstract import PipelineExtension
from cg.services.analysis_starter.configurator.file_creators.nextflow.gene_panel import (
    GenePanelFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.managed_variants import (
    ManagedVariantsFileCreator,
)


class RarediseaseExtension(PipelineExtension):
    """Contains Raredisease specific file creations which differ from the default Nextflow flow."""

    def __init__(
        self,
        gene_panel_file_creator: GenePanelFileCreator,
        managed_variants_file_creator: ManagedVariantsFileCreator,
    ):
        self.gene_panel_file_creator = gene_panel_file_creator
        self.managed_variants_file_creator = managed_variants_file_creator

    def configure(self, case_id: str, case_path: Path) -> None:
        """Perform pipeline specific actions."""
        self.gene_panel_file_creator.create(case_id=case_id, case_path=case_path)
        self.managed_variants_file_creator.create(case_id=case_id, case_path=case_path)



import logging
import re
from pathlib import Path

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.constants import FileExtensions, SequencingFileTag
from cg.constants.subject import PlinkPhenotypeStatus, PlinkSex
from cg.io.csv import write_csv
from cg.io.gzip import read_gzip_first_line
from cg.meta.workflow.fastq import is_undetermined_in_path
from cg.models.fastq import FastqFileMeta, GetFastqFileMeta
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.abstract import (
    NextflowSampleSheetCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.models import (
    RarediseaseSampleSheetEntry,
    RarediseaseSampleSheetHeaders,
)
from cg.store.models import Case, CaseSample, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class RarediseaseSampleSheetCreator(NextflowSampleSheetCreator):

    def __init__(self, housekeeper_api: HousekeeperAPI, lims: LimsAPI, store: Store):
        self.housekeeper_api = housekeeper_api
        self.lims = lims
        self.store = store

    @staticmethod
    def get_file_path(case_id: str, case_path: Path) -> Path:
        """Return the path to the params file."""
        return Path(case_path, f"{case_id}_sample_sheet").with_suffix(FileExtensions.CSV)

    def create(self, case_id: str, case_path: Path) -> None:
        """Create the sample sheet for a case."""
        file_path: Path = self.get_file_path(case_id=case_id, case_path=case_path)
        content: list[list[str]] = self._get_content(case_id)
        write_csv(file_path=file_path, content=content)

    def _get_content(self, case_id: str) -> list[list[str]]:
        """Return formatted information required to build a sample sheet for a case.
        This contains information for all samples linked to the case."""
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        sample_sheet_content: list[list[str]] = [RarediseaseSampleSheetHeaders.list()]
        for link in case.links:
            sample_sheet_content.extend(self._get_sample_sheet_content_per_sample(case_sample=link))
        return sample_sheet_content

    def _get_sample_sheet_content_per_sample(self, case_sample: CaseSample) -> list[list[str]]:
        """Collect and format information required to build a sample sheet for a single sample."""
        fastq_forward_read_paths, fastq_reverse_read_paths = self._get_paired_read_paths(
            sample=case_sample.sample
        )
        sample_sheet_entry = RarediseaseSampleSheetEntry(
            name=case_sample.sample.internal_id,
            fastq_forward_read_paths=fastq_forward_read_paths,
            fastq_reverse_read_paths=fastq_reverse_read_paths,
            sex=self._get_sex_code(case_sample.sample.sex),
            phenotype=self._get_phenotype_code(case_sample.status),
            paternal_id=case_sample.get_paternal_sample_id,
            maternal_id=case_sample.get_maternal_sample_id,
            case_id=case_sample.case.internal_id,
        )
        return sample_sheet_entry.reformat_sample_content

    def _get_paired_read_paths(self, sample: Sample) -> tuple[list[str], list[str]]:
        """Returns a tuple of paired fastq file paths for the forward and reverse read."""
        sample_metadata: list[FastqFileMeta] = self._get_fastq_metadata_for_sample(sample)
        fastq_forward_read_paths: list[str] = self._extract_read_files(
            metadata=sample_metadata, forward_read=True
        )
        fastq_reverse_read_paths: list[str] = self._extract_read_files(
            metadata=sample_metadata, reverse_read=True
        )
        return fastq_forward_read_paths, fastq_reverse_read_paths

    def _get_fastq_metadata_for_sample(self, sample: Sample) -> list[FastqFileMeta]:
        """Return FASTQ metadata objects for all fastq files linked to a sample."""
        return [
            self._parse_fastq_data(hk_file.full_path)
            for hk_file in self.housekeeper_api.files(
                bundle=sample.internal_id, tags={SequencingFileTag.FASTQ}
            )
        ]

    def _parse_fastq_data(self, fastq_path: Path) -> FastqFileMeta:
        header_line: str = read_gzip_first_line(file_path=fastq_path)
        fastq_file_meta: FastqFileMeta = self._parse_fastq_header(header_line)
        fastq_file_meta.path = fastq_path
        fastq_file_meta.undetermined = is_undetermined_in_path(fastq_path)
        matches = re.findall(r"-l[1-9]t([1-9]{2})_", str(fastq_path))
        if len(matches) > 0:
            fastq_file_meta.flow_cell_id = f"{fastq_file_meta.flow_cell_id}-{matches[0]}"
        return fastq_file_meta

    @staticmethod
    def _parse_fastq_header(line: str) -> FastqFileMeta | None:
        """Parse and return fastq header metadata.
        Handle Illumina's two different header formats
        @see https://en.wikipedia.org/wiki/FASTQ_format
        Raise:
            TypeError if unable to split line into expected parts.
        """
        parts = line.split(":")
        try:
            return GetFastqFileMeta.header_format.get(len(parts))(parts=parts)
        except TypeError as exception:
            LOG.error(f"Could not parse header format for header: {line}")
            raise exception

    @staticmethod
    def _extract_read_files(
        metadata: list[FastqFileMeta], forward_read: bool = False, reverse_read: bool = False
    ) -> list[str]:
        """Extract a list of fastq file paths for either forward or reverse reads."""
        if forward_read and not reverse_read:
            read_direction = 1
        elif reverse_read and not forward_read:
            read_direction = 2
        else:
            raise ValueError("Either forward or reverse needs to be specified")
        sorted_metadata: list = sorted(metadata, key=lambda k: k.path)
        return [
            fastq_file.path
            for fastq_file in sorted_metadata
            if fastq_file.read_direction == read_direction
        ]

    @staticmethod
    def _get_phenotype_code(phenotype: str) -> int:
        """Return Raredisease phenotype code."""
        try:
            code = PlinkPhenotypeStatus[phenotype.upper()]
        except KeyError:
            raise ValueError(f"{phenotype} is not a valid phenotype")
        return code

    @staticmethod
    def _get_sex_code(sex: str) -> PlinkSex:
        """Return Raredisease sex code."""
        try:
            code = PlinkSex[sex.upper()]
        except KeyError:
            raise ValueError(f"{sex} is not a valid sex")
        return code



from cg.constants import Workflow

WORKFLOW_LAUNCH_COMMAND_MAP: dict[Workflow, str] = {
    Workflow.MICROSALT: "{conda_binary} run --name {environment} "
    "{binary} analyse {config_file} --input {fastq_directory}"
}

WORKFLOW_VERSION_COMMAND_MAP: dict[Workflow, str] = {
    Workflow.MICROSALT: "{conda_binary} run {binary} --version"
}



from cg.constants import Workflow
from cg.store.models import Case


def get_caesar_delivery_path(case: Case) -> str:
    ticket_id: str = case.latest_ticket
    customer_id: str = case.customer.internal_id
    return f"/home/{customer_id}/inbox/{ticket_id}"


def get_scout_link(case: Case) -> str:
    customer_id: str = case.customer.internal_id
    case_name: str = case.name
    url: str = (
        "https://scout38.sys.scilifelab.se/"
        if case.data_analysis == Workflow.NALLO
        else "https://scout.scilifelab.se/"
    )
    return f"{url}{customer_id}/{case_name}"


def get_scout_links_row_separated(cases: list[Case]) -> str:
    scout_links: list[str] = [get_scout_link(case) for case in cases]
    return "\n".join(scout_links)


def get_pangolin_delivery_path(case: Case) -> str:
    customer_id: str = case.customer.internal_id
    return f"/home/{customer_id}/inbox/wwLab_automatisk_hamtning"


def get_batch_id_for_case(case: Case) -> str:
    return case.name.split("-")[1]


def get_statina_link(batch_id: str) -> str:
    return f"https://statina.clinicalgenomics.se/batches/{batch_id}"



from typing import Any

from pydantic.v1 import ValidationError

from cg.apps.lims import LimsAPI
from cg.constants.constants import CustomerId
from cg.constants.invoice import CostCenters
from cg.constants.priority import PriorityTerms
from cg.constants.sequencing import RecordType
from cg.models.invoice.invoice import (
    InvoiceApplication,
    InvoiceContact,
    InvoiceInfo,
    InvoiceReport,
)
from cg.server.ext import FlaskLims
from cg.server.ext import lims as genologics_lims
from cg.store.models import Customer, Invoice, Pool, Sample, User
from cg.store.store import Store


class InvoiceAPI:
    def __init__(self, db: Store, lims_api: LimsAPI, invoice_obj: Invoice):
        self.db = db
        self.lims_api = lims_api
        self.log = []
        self.invoice_obj = invoice_obj
        self.customer_obj = invoice_obj.customer
        self.record_type = ""
        self.raw_records = []
        self._set_record_type()
        self.genologics_lims: FlaskLims = genologics_lims
        self.invoice_info: InvoiceInfo | None = None

    def _set_record_type(self):
        """Define the record_type based on the invoice object.
        It can only be either pool or sample."""
        if self.invoice_obj.pools:
            self.raw_records = self.invoice_obj.pools
            self.record_type = RecordType.Pool
        elif self.invoice_obj.samples:
            self.record_type = RecordType.Sample
            self.raw_records = self.invoice_obj.samples

    def get_customer_by_cost_center(self, cost_center: str) -> Customer | str:
        """Return the costumer based on cost center."""
        return (
            self.db.get_customer_by_internal_id(customer_internal_id=CustomerId.CUST999)
            if cost_center.lower() == CostCenters.kth
            else self.customer_obj
        )

    def get_customer_invoice_contact(self, customer: Customer, msg: str) -> Any:
        """Return the customer invoice contact."""
        if not customer.invoice_contact:
            self.log.append(msg)
            return None
        return customer.invoice_contact

    def get_contact(
        self, customer: Customer, customer_invoice_contact: User | None, msg: str
    ) -> InvoiceContact | None:
        """Return the contact information."""
        try:
            contact = InvoiceContact(
                name=customer_invoice_contact.name,
                email=customer_invoice_contact.email,
                customer_name=customer.name,
                reference=customer.invoice_reference,
                address=customer.invoice_address,
            )
            return contact

        except ValidationError:
            self.log.append(msg)
            return None

    def get_contact_info(self, cost_center: str) -> InvoiceContact | None:
        """Return contact info for a customer."""

        msg = (
            f"Could not open/generate invoice. Contact information missing in database for "
            f"customer {self.customer_obj.internal_id}. See log files."
        )
        customer = self.get_customer_by_cost_center(cost_center=cost_center)
        customer_invoice_contact = self.get_customer_invoice_contact(customer=customer, msg=msg)
        return self.get_contact(
            customer_invoice_contact=customer_invoice_contact, customer=customer, msg=msg
        )

    def get_invoice_report(self, cost_center: str) -> dict | None:
        """Return invoice information as dictionary to generate Excel report."""

        records: list[dict] = []
        pooled_samples = []

        for raw_record in self.raw_records:
            if self.record_type == RecordType.Pool:
                pooled_samples += self.genologics_lims.samples_in_pools(
                    raw_record.name, raw_record.ticket
                )
            record = self.get_invoice_entity_record(
                cost_center=cost_center.lower(),
                discount=self.invoice_obj.discount,
                record=raw_record,
            )
            if record:
                records.append(record.dict())
            else:
                return None

        customer_obj = self.invoice_obj.customer
        contact = self.get_contact_info(cost_center)
        if not contact:
            return None

        try:
            invoice_report = InvoiceReport(
                cost_center=cost_center,
                project_number=getattr(customer_obj, f"project_account_{cost_center.lower()}"),
                customer_id=customer_obj.internal_id,
                customer_name=customer_obj.name,
                agreement=customer_obj.agreement_registration,
                invoice_id=self.invoice_obj.id,
                contact=contact.dict(),
                records=records,
                pooled_samples=pooled_samples,
                record_type=self.record_type,
            )
            return invoice_report.dict()
        except ValidationError:
            self.log.append("ValidationError in InvoiceReport class.")
            return None

    def _discount_price(self, record: Sample or Pool, discount: int = 0) -> int | None:
        """Return discount price for a sample or pool."""
        priority = self.get_priority(record, for_discount_price=True)
        full_price = getattr(record.application_version, f"price_{priority}")
        discount_factor = 1 - discount / 100
        return round(full_price * discount_factor) if full_price else 0

    def _cost_center_split_factor(
        self, price: int, cost_center: str, percent_kth: int, tag: str, version: str
    ) -> int | None:
        """Return split price based on cost center."""
        if price:
            try:
                if cost_center == CostCenters.kth:
                    split_factor = percent_kth / 100
                else:
                    split_factor = 1 - percent_kth / 100
                split_price = round(price * split_factor)
            except ValueError:
                self.log.append(
                    f"Could not calculate price for samples with application "
                    f"tag/version: {tag}/{version}. Missing %KTH"
                )
                return None
        else:
            self.log.append(
                f"Could not get price for samples with application tag/version: {tag}/{version}."
            )
            return 0
        return split_price

    def get_invoice_entity_record(
        self, cost_center: str, discount: int, record: Sample or Pool
    ) -> InvoiceInfo:
        """Return invoice information for a specific sample or pool."""
        application = self.get_application(record=record, discount=discount)
        split_discounted_price = self._cost_center_split_factor(
            price=application.discounted_price,
            cost_center=cost_center,
            percent_kth=application.percent_kth,
            tag=application.tag,
            version=application.version,
        )
        invoice_info = self.get_invoice_info(
            record=record,
            split_discounted_price=split_discounted_price,
            cost_center=cost_center,
            application=application,
        )

        return invoice_info

    def get_application(self, record: Sample or Pool, discount: int) -> InvoiceApplication | None:
        """Return the application information."""
        try:
            application = InvoiceApplication(
                tag=record.application_version.application.tag,
                version=record.application_version.version,
                percent_kth=record.application_version.application.percent_kth,
                discounted_price=self._discount_price(record, discount),
            )
            return application

        except ValidationError:
            self.log.append(f"Application tag/version seems to be missing for sample {record.id}.")

            return None

    def get_ticket(self, record: Sample or Pool) -> str:
        """Return ticket."""
        return record.ticket if self.record_type == RecordType.Pool else record.original_ticket

    def get_lims_id(self, record: Sample or Pool) -> str:
        """Return Lims id."""
        return None if self.record_type == RecordType.Pool else record.internal_id

    def get_priority(self, record: Sample or Pool, for_discount_price: bool = False) -> str:
        """Return the priority."""
        if self.customer_obj.internal_id == CustomerId.CUST032:
            priority = PriorityTerms.STANDARD
        elif self.record_type == RecordType.Pool or (
            record.priority_human == "clinical trials" and for_discount_price
        ):
            priority = PriorityTerms.RESEARCH
        else:
            priority = record.priority_human
        return priority

    def get_invoice_info(
        self, record, split_discounted_price: int, cost_center: str, application: InvoiceApplication
    ) -> InvoiceInfo:
        """Return the invoice_info to be used in the invoice report."""
        order = record.order
        ticket = self.get_ticket(record)
        lims_id = self.get_lims_id(record)
        priority = self.get_priority(record)
        price_kth: int = 0
        if cost_center == CostCenters.ki:
            price_kth: int = self._cost_center_split_factor(
                price=application.discounted_price,
                cost_center=CostCenters.kth,
                percent_kth=application.percent_kth,
                tag=application.tag,
                version=application.version,
            )

        try:
            invoice_info = InvoiceInfo(
                name=record.name,
                lims_id=lims_id,
                id=record.id,
                application_tag=record.application_version.application.tag,
                project=f"{order or 'NA'} ({ticket or 'NA'})",
                date=record.received_at.date() if record.received_at else "",
                price=split_discounted_price,
                priority=priority,
                price_kth=price_kth if cost_center == CostCenters.ki else None,
                total_price=application.discounted_price if cost_center == CostCenters.ki else None,
            )
            self.set_invoice_info(invoice_info=invoice_info)
            return invoice_info

        except ValidationError:
            self.log.append("ValidationError in InvoiceInfo class.")

    def set_invoice_info(self, invoice_info: InvoiceInfo):
        """Set invoice_info."""
        self.invoice_info = invoice_info

    def total_price(self) -> float | None:
        """Return the total price for all records in the invoice."""

        discount = self.invoice_obj.discount
        total_price = 0

        for record in self.raw_records:
            discount_price = self._discount_price(record, discount)
            if discount_price:
                total_price += discount_price
            else:
                return None
        return total_price



from enum import StrEnum

from cg.models.nf_analysis import NextflowSampleSheetEntry


class RarediseaseSampleSheetEntry(NextflowSampleSheetEntry):
    """Raredisease sample model is used when building the sample sheet."""

    phenotype: int
    sex: int
    paternal_id: str
    maternal_id: str
    case_id: str

    @property
    def reformat_sample_content(self) -> list[list[str]]:
        """
        Reformat sample sheet content as a list of lists, where each list represents a
        line in the final file.
        """
        return [
            [
                self.name,
                lane + 1,
                forward_path,
                reverse_path,
                self.sex,
                self.phenotype,
                self.paternal_id,
                self.maternal_id,
                self.case_id,
            ]
            for lane, (forward_path, reverse_path) in enumerate(
                zip(self.fastq_forward_read_paths, self.fastq_reverse_read_paths)
            )
        ]


class RarediseaseSampleSheetHeaders(StrEnum):
    sample: str = "sample"
    lane: str = "lane"
    fastq_1: str = "fastq_1"
    fastq_2: str = "fastq_2"
    sex: str = "sex"
    phenotype: str = "phenotype"
    paternal_id: str = "paternal_id"
    maternal_id: str = "maternal_id"
    case_id: str = "case_id"

    @classmethod
    def list(cls) -> list[str]:
        return list(map(lambda header: header.value, cls))



from pydantic import BaseModel, BeforeValidator, field_validator
from typing_extensions import Annotated

from cg.constants import NA_FIELD, RIN_MAX_THRESHOLD, RIN_MIN_THRESHOLD
from cg.models.delivery_report.validators import (
    get_float_as_percentage,
    get_initial_qc_as_string,
    get_number_as_string,
    get_report_string,
    get_sex_as_string,
)


class SampleMetadataModel(BaseModel):
    """
    Metrics and trending data model associated to a specific sample.

    Attributes:
        duplicates: fraction of mapped sequence that is marked as duplicate; source: workflow
        million_read_pairs: number of million read pairs obtained; source: StatusDB/sample/reads (/2*10^6)
        initial_qc: initial QC protocol flag; source: LIMS
    """

    duplicates: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    million_read_pairs: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    initial_qc: Annotated[str, BeforeValidator(get_initial_qc_as_string)] = NA_FIELD


class MipDNASampleMetadataModel(SampleMetadataModel):
    """Metrics and trending data model associated to a specific MIP DNA sample.

    Attributes:
        bait_set: panel bed used for the analysis; source: LIMS
        mapped_reads: percentage of reads aligned to the reference sequence; source: workflow
        mean_target_coverage: mean coverage of a target region; source: workflow
        pct_10x: percent of targeted bases that are covered to 10X coverage or more; source: workflow
        sex: sex predicted by the workflow; source: workflow
    """

    bait_set: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    mapped_reads: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    mean_target_coverage: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    pct_10x: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    sex: Annotated[str, BeforeValidator(get_sex_as_string)] = NA_FIELD


class NalloSampleMetadataModel(SampleMetadataModel):
    """Metrics and trending data model associated to a specific Nallo sample.

    Attributes:
        avg_sequence_length: average length of the sequenced reads; source: workflow
        coverage_bases: total number of bases aligned to the genome; source: workflow
        mean_target_coverage: mean coverage of a target region; source: Chanjo2
        median_coverage: median coverage of mapped sequence in bases; source: workflow
        pct_10x: percent of targeted bases that are covered to 10X coverage or more; source: Chanjo2
        sex: sex predicted by the workflow; source: workflow
    """

    avg_sequence_length: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    coverage_bases: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    mean_target_coverage: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    median_coverage: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    pct_10x: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    sex: Annotated[str, BeforeValidator(get_sex_as_string)] = NA_FIELD


class RarediseaseSampleMetadataModel(SampleMetadataModel):
    """Metrics and trending data model associated to a specific MIP DNA sample.

    Attributes:
        bait_set: panel bed used for the analysis; source: LIMS
        mapped_reads: percentage of reads aligned to the reference sequence; source: workflow
        mean_target_coverage: mean coverage of a target region; source: Chanjo2
        pct_10x: percent of targeted bases that are covered to 10X coverage or more; source: Chanjo2
        sex: sex predicted by the workflow; source: workflow
    """

    bait_set: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    mapped_reads: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    mean_target_coverage: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    pct_10x: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    sex: Annotated[str, BeforeValidator(get_sex_as_string)] = NA_FIELD


class BalsamicSampleMetadataModel(SampleMetadataModel):
    """Metrics and trending data model associated to a specific BALSAMIC sample.

    Attributes:
        mean_insert_size: mean insert size of the distribution; source: workflow
        fold_80: fold 80 base penalty; source: workflow
    """

    mean_insert_size: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    fold_80: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD


class BalsamicTargetedSampleMetadataModel(BalsamicSampleMetadataModel):
    """Metrics and trending data model associated to a specific BALSAMIC sample.

    Attributes:
        bait_set: panel bed used for the analysis; source: LIMS
        bait_set_version: panel bed version; source: workflow
        median_target_coverage: median coverage of a target region in bases; source: workflow
        pct_250x: percent of targeted bases that are covered to 250X coverage or more; source: workflow
        pct_500x: percent of targeted bases that are covered to 500X coverage or more; source: workflow
    """

    bait_set: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    bait_set_version: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    gc_dropout: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    median_target_coverage: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    pct_250x: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    pct_500x: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD


class BalsamicWGSSampleMetadataModel(BalsamicSampleMetadataModel):
    """Metrics and trending data model associated to a specific BALSAMIC sample.

    Attributes:
        median_coverage: median coverage in bases of the genome territory; source: workflow
        pct_15x: fraction of bases that attained at least 15X sequence coverage; source: workflow
        pct_60x: fraction of bases that attained at least 15X sequence coverage; source: workflow
        pct_reads_improper_pairs: fraction of reads that are not properly aligned in pairs
    """

    median_coverage: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    pct_15x: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    pct_60x: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    pct_reads_improper_pairs: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD


class SequencingSampleMetadataModel(SampleMetadataModel):
    """Metrics and trending data model associated to the sequencing of a sample."

    Attributes:
        gc_content: percentage of GC bases calculated on trimmed reads; source: workflow
        mean_length_r1: average length of reads that pass QC filters; source: workflow
    """

    gc_content: Annotated[str, BeforeValidator(get_float_as_percentage)] = NA_FIELD
    mean_length_r1: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD


class WTSSampleMetadataModel(SequencingSampleMetadataModel):
    """Metrics and trending data model associated to a WHOLE_TRANSCRIPTOME_SEQUENCING sample.

    Attributes:
        bias_5_3: bias is the ratio between read counts; source: workflow
        dv200: percentage of RNA fragments > 200 nucleotides; source: LIMS
        input_amount: input amount in ng; source: LIMS
        mrna_bases:  proportion of bases that originate from messenger RNA; source: workflow
        pct_adapter: proportion of reads that contain adapter sequences; source: workflow
        pct_surviving: percentage of reads that pass quality control filters; source: workflow
        q20_rate: proportion of bases with a minimum Phred score of 20; source: workflow
        q30_rate: proportion of bases with a minimum Phred score of 30; source: workflow
        ribosomal_bases: proportion of bases that originate from ribosomal RNA; source: workflow
        rin: RNA integrity number; source: LIMS
        uniquely_mapped_reads: percentage of mapped reads; source: workflow
    """

    bias_5_3: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    dv200: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    input_amount: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    mrna_bases: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    pct_adapter: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    pct_surviving: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    q20_rate: Annotated[str, BeforeValidator(get_float_as_percentage)] = NA_FIELD
    q30_rate: Annotated[str, BeforeValidator(get_float_as_percentage)] = NA_FIELD
    ribosomal_bases: Annotated[str, BeforeValidator(get_float_as_percentage)] = NA_FIELD
    rin: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    uniquely_mapped_reads: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD

    @field_validator("rin")
    @classmethod
    def ensure_rin_thresholds(cls, rin: str) -> str:
        if rin != NA_FIELD:
            rin_number = float(rin)
            if RIN_MIN_THRESHOLD <= rin_number <= RIN_MAX_THRESHOLD:
                return str(rin_number)
        return NA_FIELD


class RnafusionSampleMetadataModel(WTSSampleMetadataModel):
    """Metrics and trending data model associated to a specific Rnafusion sample.

    Attributes:
        insert_size: distance between paired-end sequencing reads in a DNA fragment
        insert_size_peak: insert size length; source: workflow
        mapped_reads: percentage of reads aligned to the reference sequence; source: workflow
    """

    insert_size: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    insert_size_peak: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    mapped_reads: Annotated[str, BeforeValidator(get_float_as_percentage)] = NA_FIELD


class TaxprofilerSampleMetadataModel(SequencingSampleMetadataModel):
    """Metrics and trending data model associated to a Taxprofiler sample.

    Attributes:
        average_read_length: average length of reads; source: workflow
        mapped_reads: reads aligned to the reference sequence; source: workflow
        mean_length_r2: average length of reads for read2; source: workflow
        million_read_pairs_after_filtering: number of reads after filtering; source: workflow
    """

    average_read_length: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    mapped_reads: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    mean_length_r2: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    million_read_pairs_after_filtering: Annotated[str, BeforeValidator(get_number_as_string)] = (
        NA_FIELD
    )


class TomteSampleMetadataModel(WTSSampleMetadataModel):
    """Metrics and trending data model associated to a Tomte sample.

    Attributes:
        pct_intergenic_bases: proportion of genomic bases located between genes; source: workflow
        pct_intronic_bases: proportion of genomic bases within intronic regions; source: workflow
    """

    pct_intergenic_bases: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD
    pct_intronic_bases: Annotated[str, BeforeValidator(get_number_as_string)] = NA_FIELD



from datetime import date

from pydantic import BeforeValidator, PrivateAttr, field_serializer, model_validator
from typing_extensions import Annotated

from cg.constants.orderforms import ORIGINAL_LAB_ADDRESSES, REGION_CODES
from cg.models.orders.sample_base import ControlEnum, PriorityEnum
from cg.services.orders.validation.constants import ElutionBuffer, ExtractionMethod
from cg.services.orders.validation.models.sample import Sample
from cg.services.orders.validation.order_types.mutant.constants import (
    OriginalLab,
    PreProcessingMethod,
    Primer,
    Region,
    SelectionCriteria,
)
from cg.services.orders.validation.utils import parse_buffer, parse_control, parse_extraction_method


class MutantSample(Sample):
    collection_date: date
    concentration_sample: float | None = None
    control: Annotated[ControlEnum, BeforeValidator(parse_control)] = ControlEnum.not_control
    elution_buffer: Annotated[ElutionBuffer, BeforeValidator(parse_buffer)]
    extraction_method: Annotated[ExtractionMethod, BeforeValidator(parse_extraction_method)]
    _lab_code: str = PrivateAttr(default="SE100 Karolinska")
    organism: str
    organism_other: str | None = None
    original_lab: OriginalLab
    original_lab_address: str
    pre_processing_method: PreProcessingMethod
    primer: Primer
    priority: PriorityEnum
    quantity: int | None = None
    reference_genome: str
    region: Region
    region_code: str
    selection_criteria: SelectionCriteria
    _verified_organism: bool | None = PrivateAttr(default=None)

    @model_validator(mode="before")
    @classmethod
    def set_original_lab_address(cls, data: any) -> any:
        if isinstance(data, dict):
            is_set = bool(data.get("original_lab_address"))
            if not is_set and data.get("original_lab"):
                data["original_lab_address"] = ORIGINAL_LAB_ADDRESSES[data["original_lab"]]
        return data

    @model_validator(mode="before")
    @classmethod
    def set_region_code(cls, data: any) -> any:
        if isinstance(data, dict):
            is_set = bool(data.get("region_code"))
            if not is_set and data.get("region"):
                data["region_code"] = REGION_CODES[data["region"]]
        return data

    @field_serializer("collection_date")
    def serialize_collection_date(self, value: date) -> str:
        return value.isoformat()

    def model_dump(self, **kwargs) -> dict:
        data = super().model_dump(**kwargs)
        data["lab_code"] = self._lab_code
        data["verified_organism"] = self._verified_organism
        return data



"""Delete observations CLI."""

import logging

import rich_click as click
from sqlalchemy.orm import Query

from cg.cli.upload.observations.utils import get_observations_api
from cg.cli.workflow.commands import ARGUMENT_CASE_ID, OPTION_LOQUSDB_SUPPORTED_WORKFLOW
from cg.constants.cli_options import DRY_RUN, SKIP_CONFIRMATION
from cg.constants.constants import Workflow
from cg.exc import CaseNotFoundError, LoqusdbError
from cg.meta.observations.balsamic_observations_api import BalsamicObservationsAPI
from cg.meta.observations.mip_dna_observations_api import MipDNAObservationsAPI
from cg.meta.observations.nallo_observations_api import NalloObservationsAPI
from cg.meta.observations.raredisease_observations_api import RarediseaseObservationsAPI
from cg.models.cg_config import CGConfig
from cg.store.store import Store

LOG = logging.getLogger(__name__)


@click.command("observations")
@ARGUMENT_CASE_ID
@SKIP_CONFIRMATION
@DRY_RUN
@click.pass_obj
def delete_observations(context: CGConfig, case_id: str, dry_run: bool, skip_confirmation: bool):
    """Delete a case from Loqusdb and reset the Loqusdb IDs in StatusDB."""
    observations_api: (
        BalsamicObservationsAPI
        | MipDNAObservationsAPI
        | NalloObservationsAPI
        | RarediseaseObservationsAPI
    ) = get_observations_api(context=context, case_id=case_id, upload=False)
    if dry_run:
        LOG.info(f"Dry run. This would delete all variants in Loqusdb for case: {case_id}")
        return
    LOG.info(f"This will delete all variants in Loqusdb for case: {case_id}")
    if skip_confirmation or click.confirm("Do you want to continue?", abort=True):
        observations_api.delete_case(case_id)


@click.command("available-observations")
@OPTION_LOQUSDB_SUPPORTED_WORKFLOW
@SKIP_CONFIRMATION
@DRY_RUN
@click.pass_context
def delete_available_observations(
    context: click.Context, workflow: Workflow | None, dry_run: bool, skip_confirmation: bool
):
    """Delete available observation from Loqusdb."""
    status_db: Store = context.obj.status_db
    uploaded_observations: Query = status_db.observations_uploaded(workflow)
    LOG.info(
        f"This would delete observations for the following cases: "
        f"{[case.internal_id for case in uploaded_observations]}"
    )
    if skip_confirmation or click.confirm("Do you want to continue?", abort=True):
        for case in uploaded_observations:
            try:
                LOG.info(f"Will delete observations for {case.internal_id}")
                context.invoke(
                    delete_observations,
                    case_id=case.internal_id,
                    dry_run=dry_run,
                    skip_confirmation=skip_confirmation,
                )
            except (CaseNotFoundError, LoqusdbError) as error:
                LOG.error(f"Error deleting observations for {case.internal_id}: {error}")
                continue



import logging

from housekeeper.store.models import Version

from cg.apps.lims import LimsAPI
from cg.apps.madeline.api import MadelineAPI
from cg.constants.housekeeper_tags import HK_DELIVERY_REPORT_TAG
from cg.constants.scout import (
    MIP_CASE_TAGS,
    MIP_SAMPLE_TAGS,
    RANK_MODEL_THRESHOLD,
    GenomeBuild,
    UploadTrack,
)
from cg.meta.upload.scout.hk_tags import CaseTags, SampleTags
from cg.meta.upload.scout.scout_config_builder import ScoutConfigBuilder
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.models.mip.mip_analysis import MipAnalysis
from cg.models.scout.scout_load_config import MipLoadConfig, ScoutMipIndividual
from cg.store.models import Analysis, CaseSample

LOG = logging.getLogger(__name__)


class MipConfigBuilder(ScoutConfigBuilder):
    def __init__(
        self,
        mip_analysis_api: MipAnalysisAPI,
        lims_api: LimsAPI,
        madeline_api: MadelineAPI,
    ):
        super().__init__(
            lims_api=lims_api,
        )
        self.case_tags: CaseTags = CaseTags(**MIP_CASE_TAGS)
        self.sample_tags: SampleTags = SampleTags(**MIP_SAMPLE_TAGS)
        self.mip_analysis_api: MipAnalysisAPI = mip_analysis_api
        self.lims_api: LimsAPI = lims_api
        self.madeline_api: MadelineAPI = madeline_api

    def build_load_config(
        self,
        hk_version: Version,
        analysis: Analysis,
        rank_score_threshold: int = RANK_MODEL_THRESHOLD,
    ) -> MipLoadConfig:
        """Create a MIP specific load config for uploading analysis to Scout."""
        LOG.info("Generate load config for mip case")
        load_config: MipLoadConfig = MipLoadConfig(
            track=UploadTrack.RARE_DISEASE.value,
            delivery_report=self.get_file_from_hk(
                hk_tags={HK_DELIVERY_REPORT_TAG}, hk_version=hk_version
            ),
        )
        self.add_common_info_to_load_config(load_config=load_config, analysis=analysis)
        mip_analysis_data: MipAnalysis = self.mip_analysis_api.get_latest_metadata(
            analysis.case.internal_id
        )
        load_config.human_genome_build = (
            GenomeBuild.hg38
            if GenomeBuild.hg38 in mip_analysis_data.genome_build
            else GenomeBuild.hg19
        )
        load_config.rank_score_threshold = rank_score_threshold
        load_config.rank_model_version = mip_analysis_data.rank_model_version
        load_config.sv_rank_model_version = mip_analysis_data.sv_rank_model_version

        load_config.gene_panels = (
            self.mip_analysis_api.get_aggregated_panels(
                customer_id=analysis.case.customer.internal_id,
                default_panels=set(analysis.case.panels),
            )
            or None
        )

        self.include_case_files(load_config=load_config, hk_version=hk_version)

        LOG.info("Building samples")
        db_sample: CaseSample
        for db_sample in analysis.case.links:
            load_config.samples.append(
                self.build_config_sample(case_sample=db_sample, hk_version=hk_version)
            )
        self.include_pedigree_picture(load_config=load_config, analysis=analysis)
        return load_config

    def include_case_files(self, load_config: MipLoadConfig, hk_version: Version) -> None:
        """Include case level files for mip case"""
        LOG.info("Including MIP specific case level files")
        load_config.peddy_check = self.get_file_from_hk(
            self.case_tags.peddy_check, hk_version=hk_version
        )
        load_config.peddy_ped = self.get_file_from_hk(
            self.case_tags.peddy_ped, hk_version=hk_version
        )
        load_config.peddy_sex = self.get_file_from_hk(
            self.case_tags.peddy_sex, hk_version=hk_version
        )
        load_config.smn_tsv = self.get_file_from_hk(self.case_tags.smn_tsv, hk_version=hk_version)
        load_config.vcf_mei = self.get_file_from_hk(self.case_tags.vcf_mei, hk_version=hk_version)
        load_config.vcf_mei_research = self.get_file_from_hk(
            self.case_tags.vcf_mei_research, hk_version=hk_version
        )
        load_config.vcf_snv = self.get_file_from_hk(self.case_tags.snv_vcf, hk_version=hk_version)
        load_config.vcf_snv_research = self.get_file_from_hk(
            self.case_tags.snv_research_vcf, hk_version=hk_version
        )
        load_config.vcf_str = self.get_file_from_hk(self.case_tags.vcf_str, hk_version=hk_version)
        load_config.vcf_sv = self.get_file_from_hk(self.case_tags.sv_vcf, hk_version=hk_version)
        load_config.vcf_sv_research = self.get_file_from_hk(
            self.case_tags.sv_research_vcf, hk_version=hk_version
        )
        self.include_multiqc_report(load_config=load_config, hk_version=hk_version)

    def include_sample_files(self, config_sample: ScoutMipIndividual, hk_version: Version) -> None:
        """Include sample level files that are optional for mip samples"""
        LOG.info("Including MIP specific sample level files")
        sample_id: str = config_sample.sample_id
        config_sample.vcf2cytosure = self.get_sample_file(
            hk_tags=self.sample_tags.vcf2cytosure,
            sample_id=sample_id,
            hk_version=hk_version,
        )
        config_sample.mt_bam = self.get_sample_file(
            hk_tags=self.sample_tags.mt_bam, sample_id=sample_id, hk_version=hk_version
        )
        config_sample.chromograph_images.autozygous = self.remove_chromosome_substring(
            file_path=self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_autozyg,
                sample_id=sample_id,
                hk_version=hk_version,
            )
        )
        config_sample.chromograph_images.coverage = self.remove_chromosome_substring(
            file_path=self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_coverage,
                sample_id=sample_id,
                hk_version=hk_version,
            )
        )
        config_sample.chromograph_images.upd_regions = self.remove_chromosome_substring(
            file_path=self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_regions,
                sample_id=sample_id,
                hk_version=hk_version,
            )
        )
        config_sample.chromograph_images.upd_sites = self.remove_chromosome_substring(
            file_path=self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_sites,
                sample_id=sample_id,
                hk_version=hk_version,
            )
        )
        config_sample.reviewer.alignment = self.get_sample_file(
            hk_tags=self.sample_tags.reviewer_alignment,
            sample_id=sample_id,
            hk_version=hk_version,
        )
        config_sample.reviewer.alignment_index = self.get_sample_file(
            hk_tags=self.sample_tags.reviewer_alignment_index,
            sample_id=sample_id,
            hk_version=hk_version,
        )
        config_sample.reviewer.vcf = self.get_sample_file(
            hk_tags=self.sample_tags.reviewer_vcf,
            sample_id=sample_id,
            hk_version=hk_version,
        )
        config_sample.reviewer.catalog = self.get_file_from_hk(
            hk_tags=self.case_tags.str_catalog, hk_version=hk_version
        )
        config_sample.mitodel_file = self.get_sample_file(
            hk_tags=self.sample_tags.mitodel_file,
            sample_id=sample_id,
            hk_version=hk_version,
        )



from pathlib import Path

from cg.apps.scout.scoutapi import ScoutAPI
from cg.constants import FileExtensions, GenePanelMasterList, Workflow
from cg.constants.gene_panel import GenePanelCombo, GenePanelGenomeBuild
from cg.io.txt import write_txt
from cg.services.analysis_starter.configurator.file_creators.nextflow.utils import get_genome_build
from cg.store.models import Case
from cg.store.store import Store


class GenePanelFileCreator:
    def __init__(self, store: Store, scout_api: ScoutAPI):
        self.store = store
        self.scout_api = scout_api

    @staticmethod
    def get_file_path(case_path: Path) -> Path:
        return Path(case_path, "gene_panels").with_suffix(FileExtensions.BED)

    def create(self, case_id: str, case_path: Path) -> None:
        file_path: Path = self.get_file_path(case_path)
        content: list[str] = self._get_content(case_id)
        write_txt(file_path=file_path, content=content)

    def _get_content(self, case_id: str) -> list[str]:
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        genome_build: GenePanelGenomeBuild = get_genome_build(workflow=Workflow(case.data_analysis))
        all_panels: list[str] = self._get_aggregated_panels(
            customer_id=case.customer.internal_id, default_panels=set(case.panels)
        )
        return self.scout_api.export_panels(build=genome_build, panels=all_panels)

    def _get_aggregated_panels(self, customer_id: str, default_panels: set[str]) -> list[str]:
        """Check if customer is collaborator for gene panel master list
        and if all default panels are included in the gene panel master list.
        If not, add gene panel combo and broad non-specific gene panels.
        Return an aggregated gene panel."""
        if GenePanelMasterList.is_customer_collaborator_and_panels_in_gene_panels_master_list(
            customer_id=customer_id, gene_panels=default_panels
        ):
            return GenePanelMasterList.get_panel_names()
        all_panels: set[str] = self._add_gene_panel_combo(gene_panels=default_panels)
        all_panels |= GenePanelMasterList.get_non_specific_gene_panels()
        return list(all_panels)

    @staticmethod
    def _add_gene_panel_combo(gene_panels: set[str]) -> set[str]:
        """
        Add gene panels combinations for gene panels being part of gene panel combination and
        return updated gene panels.
        """
        additional_panels = set()
        for panel in gene_panels:
            if panel in GenePanelCombo.COMBO_1:
                additional_panels |= GenePanelCombo.COMBO_1.get(panel)
        gene_panels |= additional_panels
        return gene_panels



import datetime as dt

from cg.apps.tb.api import TrailblazerAPI
from cg.constants.constants import Workflow
from cg.constants.tb import AnalysisStatus, AnalysisType
from cg.exc import CaseNotFoundError
from cg.meta.workflow.utils.utils import MAP_TO_TRAILBLAZER_PRIORITY
from cg.store.models import Analysis, Case
from cg.store.store import Store


class RawDataAnalysisService:
    def __init__(self, store: Store, trailblazer_api: TrailblazerAPI):
        self.store = store
        self.trailblazer_api = trailblazer_api

    def store_analysis(self, case_id: str) -> None:
        case: Case = self._get_case(case_id)
        self._add_analysis_to_store(case)
        self._add_analysis_to_trailblazer(case)

    def _get_case(self, case_id: str) -> Case:
        if case := self.store.get_case_by_internal_id(case_id):
            return case
        raise CaseNotFoundError(f"Case {case_id} not found.")

    def _add_analysis_to_store(self, case: Case) -> None:
        new_analysis: Analysis = self.store.add_analysis(
            workflow=Workflow.RAW_DATA,
            completed_at=dt.datetime.now(),
            primary=True,
            started_at=dt.datetime.now(),
            case_id=case.id,
        )
        self.store.session.add(new_analysis)
        self.store.session.commit()

    def _add_analysis_to_trailblazer(self, case: Case) -> None:
        self.trailblazer_api.add_pending_analysis(
            case_id=case.internal_id,
            analysis_type=AnalysisType.OTHER,
            config_path="",
            order_id=case.latest_order.id,
            out_dir="",
            priority=MAP_TO_TRAILBLAZER_PRIORITY[case.priority],
            workflow=Workflow.RAW_DATA,
            ticket=case.latest_ticket,
        )
        self.trailblazer_api.set_analysis_status(
            case_id=case.internal_id, status=AnalysisStatus.COMPLETED
        )



from enum import StrEnum

from cg.constants import DataDelivery


class NalloDeliveryType(StrEnum):
    ANALYSIS = DataDelivery.ANALYSIS_FILES
    ANALYSIS_SCOUT = DataDelivery.ANALYSIS_SCOUT
    NO_DELIVERY = DataDelivery.NO_DELIVERY
    SCOUT = DataDelivery.SCOUT
    RAW_DATA_ANALYSIS = DataDelivery.RAW_DATA_ANALYSIS
    RAW_DATA_ANALYSIS_SCOUT = DataDelivery.RAW_DATA_ANALYSIS_SCOUT
    RAW_DATA_SCOUT = DataDelivery.RAW_DATA_SCOUT



"""Module for the factory of the deliver files service."""

from typing import Type

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants import DataDelivery, Workflow
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.services.analysis_service.analysis_service import AnalysisService
from cg.services.deliver_files.constants import DeliveryDestination, DeliveryStructure
from cg.services.deliver_files.deliver_files_service.deliver_files_service import (
    DeliverFilesService,
)
from cg.services.deliver_files.deliver_files_service.exc import DeliveryTypeNotSupported
from cg.services.deliver_files.file_fetcher.abstract import FetchDeliveryFilesService
from cg.services.deliver_files.file_fetcher.analysis_raw_data_service import (
    RawDataAndAnalysisDeliveryFileFetcher,
)
from cg.services.deliver_files.file_fetcher.analysis_service import AnalysisDeliveryFileFetcher
from cg.services.deliver_files.file_fetcher.raw_data_service import RawDataDeliveryFileFetcher
from cg.services.deliver_files.file_formatter.destination.abstract import (
    DeliveryDestinationFormatter,
)
from cg.services.deliver_files.file_formatter.destination.base_service import BaseDeliveryFormatter
from cg.services.deliver_files.file_formatter.files.case_service import CaseFileFormatter
from cg.services.deliver_files.file_formatter.files.concatenation_service import (
    SampleFileConcatenationFormatter,
)
from cg.services.deliver_files.file_formatter.files.mutant_service import MutantFileFormatter
from cg.services.deliver_files.file_formatter.files.sample_service import (
    FileManager,
    SampleFileFormatter,
)
from cg.services.deliver_files.file_formatter.path_name.abstract import PathNameFormatter
from cg.services.deliver_files.file_formatter.path_name.flat_structure import (
    FlatStructurePathFormatter,
)
from cg.services.deliver_files.file_formatter.path_name.nested_structure import (
    NestedStructurePathFormatter,
)
from cg.services.deliver_files.file_mover.abstract import DestinationFilesMover
from cg.services.deliver_files.file_mover.base_service import BaseDestinationFilesMover
from cg.services.deliver_files.file_mover.customer_inbox_service import (
    CustomerInboxDestinationFilesMover,
)
from cg.services.deliver_files.rsync.service import DeliveryRsyncService
from cg.services.deliver_files.tag_fetcher.abstract import FetchDeliveryFileTagsService
from cg.services.deliver_files.tag_fetcher.fohm_upload_service import FOHMUploadTagsFetcher
from cg.services.deliver_files.tag_fetcher.sample_and_case_service import (
    SampleAndCaseDeliveryTagsFetcher,
)
from cg.services.deliver_files.utils import FileMover
from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)
from cg.store.models import Case
from cg.store.store import Store


class DeliveryServiceFactory:
    """
    Class to build the delivery services based on case, workflow, delivery type, delivery
    destination and delivery structure.
    The delivery destination is used to specify delivery to the customer or for external upload.
    Workflow is used to specify the workflow of the case and is required for the tag fetcher.
    Delivery type is used to specify the type of delivery to perform.
    Delivery structure is used to specify the structure of the delivery.
    """

    def __init__(
        self,
        store: Store,
        lims_api: LimsAPI,
        hk_api: HousekeeperAPI,
        rsync_service: DeliveryRsyncService,
        tb_service: TrailblazerAPI,
        analysis_service: AnalysisService,
    ):
        self.store = store
        self.lims_api = lims_api
        self.hk_api = hk_api
        self.rsync_service = rsync_service
        self.tb_service = tb_service
        self.analysis_service = analysis_service

    @staticmethod
    def _sanitise_delivery_type(delivery_type: DataDelivery) -> DataDelivery:
        """Sanitise the delivery type.
        We have multiple delivery types that are a combination of other delivery types or uploads.
        Here we make sure to convert unsupported delivery types to supported ones.
        args:
            delivery_type: The type of delivery to perform.
        """
        if delivery_type in [DataDelivery.FASTQ_QC, DataDelivery.FASTQ_SCOUT]:
            return DataDelivery.FASTQ
        if delivery_type == DataDelivery.ANALYSIS_SCOUT:
            return DataDelivery.ANALYSIS_FILES
        if delivery_type in [
            DataDelivery.FASTQ_ANALYSIS_SCOUT,
            DataDelivery.FASTQ_QC_ANALYSIS,
        ]:
            return DataDelivery.FASTQ_ANALYSIS
        if delivery_type == DataDelivery.RAW_DATA_SCOUT:
            return DataDelivery.BAM
        if delivery_type == DataDelivery.RAW_DATA_ANALYSIS_SCOUT:
            return DataDelivery.RAW_DATA_ANALYSIS
        return delivery_type

    @staticmethod
    def _validate_delivery_type(delivery_type: DataDelivery) -> None:
        """
        Check if the delivery type is supported. Raises DeliveryTypeNotSupported error.
        args:
            delivery_type: The type of delivery to perform.
        Raises:
            DeliveryTypeNotSupported: If the delivery type is not supported.
        """
        if delivery_type not in [
            DataDelivery.ANALYSIS_FILES,
            DataDelivery.BAM,
            DataDelivery.FASTQ,
            DataDelivery.FASTQ_ANALYSIS,
            DataDelivery.RAW_DATA_ANALYSIS,
        ]:
            raise DeliveryTypeNotSupported(
                f"Delivery type {delivery_type} is not supported. Supported delivery types are"
                f" {DataDelivery.FASTQ}, {DataDelivery.ANALYSIS_FILES},"
                f" {DataDelivery.FASTQ_ANALYSIS}, {DataDelivery.BAM}, {DataDelivery.RAW_DATA_ANALYSIS}."
            )

    @staticmethod
    def _get_file_tag_fetcher(
        delivery_destination: DeliveryDestination,
    ) -> FetchDeliveryFileTagsService:
        """
        Return the correct file tag fetcher based on the delivery destination.
        NOTE: The FOHM upload requires a special tag system, all the other destination use
        sample and case tags.
        args:
            delivery_destination: The destination of the delivery defaults to customer.
        """
        if delivery_destination == DeliveryDestination.FOHM:
            return FOHMUploadTagsFetcher()
        return SampleAndCaseDeliveryTagsFetcher()

    def _get_file_fetcher(
        self, delivery_type: DataDelivery, delivery_destination: DeliveryDestination
    ) -> FetchDeliveryFilesService:
        """Get the file fetcher based on the delivery type.
        args:
            delivery_type: The type of delivery to perform.
            delivery_destination: The destination of the delivery defaults to customer. See DeliveryDestination enum for explanation.

        """
        service_map: dict[DataDelivery, Type[FetchDeliveryFilesService]] = {
            DataDelivery.FASTQ: RawDataDeliveryFileFetcher,
            DataDelivery.ANALYSIS_FILES: AnalysisDeliveryFileFetcher,
            DataDelivery.FASTQ_ANALYSIS: RawDataAndAnalysisDeliveryFileFetcher,
            DataDelivery.RAW_DATA_ANALYSIS: RawDataAndAnalysisDeliveryFileFetcher,
            DataDelivery.BAM: RawDataDeliveryFileFetcher,
        }
        file_tag_fetcher: FetchDeliveryFileTagsService = self._get_file_tag_fetcher(
            delivery_destination=delivery_destination
        )
        return service_map[delivery_type](
            status_db=self.store,
            hk_api=self.hk_api,
            tags_fetcher=file_tag_fetcher,
        )

    def _convert_workflow(self, case: Case) -> Workflow:
        """Change the workflow of a Microbial Fastq case to Microsalt to allow the concatenation of fastq files.
        With the introduction of the microbial-fastq delivery type, an unsupported combination of delivery type and
        workflow setup is required. This function makes sure that a raw data workflow with microbial fastq delivery
        type is treated as a microsalt workflow so that the microbial-fastq sample files can be concatenated.
        args:
            case: The case to convert the workflow for
        """
        tag: str = case.samples[0].application_version.application.tag
        microbial_tags: list[str] = [
            application.tag
            for application in self.store.get_active_applications_by_prep_category(
                prep_category=SeqLibraryPrepCategory.MICROBIAL
            )
        ]
        if case.data_analysis == Workflow.RAW_DATA and tag in microbial_tags:
            return Workflow.MICROSALT
        return case.data_analysis

    def _get_sample_file_formatter(
        self,
        case: Case,
        delivery_structure: DeliveryStructure = DeliveryStructure.NESTED,
    ) -> SampleFileFormatter | SampleFileConcatenationFormatter | MutantFileFormatter:
        """Return the appropriate file formatter service based on the workflow.
        Depending on the delivery structure the path name formatter will be different.
        Args:
            case: The case to deliver files for.
            delivery_structure: The structure of the delivery. See DeliveryStructure enum for explanation. Defaults to nested.
        """

        converted_workflow: Workflow = self._convert_workflow(case)
        if converted_workflow in [Workflow.MICROSALT]:
            return SampleFileConcatenationFormatter(
                file_manager=FileManager(),
                path_name_formatter=self._get_path_name_formatter(delivery_structure),
                concatenation_service=FastqConcatenationService(),
            )
        if converted_workflow == Workflow.MUTANT:
            return MutantFileFormatter(
                lims_api=self.lims_api,
                file_manager=FileManager(),
                file_formatter=SampleFileConcatenationFormatter(
                    file_manager=FileManager(),
                    path_name_formatter=self._get_path_name_formatter(delivery_structure),
                    concatenation_service=FastqConcatenationService(),
                ),
            )
        return SampleFileFormatter(
            file_manager=FileManager(),
            path_name_formatter=self._get_path_name_formatter(delivery_structure),
        )

    def _get_case_file_formatter(self, delivery_structure: DeliveryStructure) -> CaseFileFormatter:
        """
        Get the case file formatter based on the delivery structure.
        args:
            delivery_structure: The structure of the delivery. See DeliveryStructure enum for explanation.
        """
        return CaseFileFormatter(
            file_manager=FileManager(),
            path_name_formatter=self._get_path_name_formatter(delivery_structure),
        )

    @staticmethod
    def _get_path_name_formatter(
        delivery_structure: DeliveryStructure,
    ) -> PathNameFormatter:
        """
        Get the path name formatter based on the delivery destination
        args:
            delivery_structure: The structure of the delivery. See DeliveryStructure enum for explanation.
        """
        if delivery_structure == DeliveryStructure.FLAT:
            return FlatStructurePathFormatter()
        return NestedStructurePathFormatter()

    @staticmethod
    def _get_file_mover(
        delivery_destination: DeliveryDestination,
    ) -> CustomerInboxDestinationFilesMover | BaseDestinationFilesMover:
        """Get the file mover based on the delivery type.
        args:
            delivery_destination: The destination of the delivery. See DeliveryDestination enum for explanation.
        """
        if delivery_destination in [DeliveryDestination.BASE, DeliveryDestination.FOHM]:
            return BaseDestinationFilesMover(FileMover(FileManager()))
        return CustomerInboxDestinationFilesMover(FileMover(FileManager()))

    def _get_file_formatter(
        self,
        delivery_structure: DeliveryStructure,
        case: Case,
    ) -> DeliveryDestinationFormatter:
        """
        Get the file formatter service based on the delivery destination.
        args:
            delivery_structure: The structure of the delivery. See DeliveryStructure enum for explanation.
            case: The case to deliver files for.
        """
        sample_file_formatter: (
            SampleFileFormatter | SampleFileConcatenationFormatter | MutantFileFormatter
        ) = self._get_sample_file_formatter(case=case, delivery_structure=delivery_structure)
        case_file_formatter: CaseFileFormatter = self._get_case_file_formatter(
            delivery_structure=delivery_structure
        )
        return BaseDeliveryFormatter(
            case_file_formatter=case_file_formatter,
            sample_file_formatter=sample_file_formatter,
        )

    def build_delivery_service(
        self,
        case: Case,
        delivery_type: DataDelivery | None = None,
        delivery_destination: DeliveryDestination = DeliveryDestination.CUSTOMER,
        delivery_structure: DeliveryStructure = DeliveryStructure.NESTED,
    ) -> DeliverFilesService:
        """Build a delivery service for a case.
        args:
            case: The case to deliver files for.
            delivery_type: The type of delivery to perform. See DataDelivery enum for explanation.
            delivery_destination: Representation of where the files will be uploaded.
                Defaults to 'customer'. See DeliveryDestination enum for explanation.
            delivery_structure: File structure for the delivery directory. Defaults to 'nested'.
                See DeliveryStructure enum for explanation.
        """
        delivery_type: DataDelivery = self._sanitise_delivery_type(
            delivery_type if delivery_type else case.data_delivery
        )
        self._validate_delivery_type(delivery_type)
        file_fetcher: FetchDeliveryFilesService = self._get_file_fetcher(
            delivery_type=delivery_type, delivery_destination=delivery_destination
        )
        file_move_service: DestinationFilesMover = self._get_file_mover(
            delivery_destination=delivery_destination
        )
        file_formatter: DeliveryDestinationFormatter = self._get_file_formatter(
            case=case, delivery_structure=delivery_structure
        )
        return DeliverFilesService(
            delivery_file_manager_service=file_fetcher,
            move_file_service=file_move_service,
            file_formatter_service=file_formatter,
            status_db=self.store,
            rsync_service=self.rsync_service,
            tb_service=self.tb_service,
            analysis_service=self.analysis_service,
        )



from pathlib import Path

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.constants import Workflow
from cg.constants.nextflow import NEXTFLOW_WORKFLOWS
from cg.meta.workflow.fastq import MicrosaltFastqHandler
from cg.models.cg_config import CGConfig, CommonAppConfig
from cg.services.analysis_starter.configurator.configurator import Configurator
from cg.services.analysis_starter.configurator.extensions.abstract import PipelineExtension
from cg.services.analysis_starter.configurator.extensions.raredisease import RarediseaseExtension
from cg.services.analysis_starter.configurator.file_creators.microsalt_config import (
    MicrosaltConfigFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.config_file import (
    NextflowConfigFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.gene_panel import (
    GenePanelFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.managed_variants import (
    ManagedVariantsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.abstract import (
    ParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.raredisease import (
    RarediseaseParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.abstract import (
    NextflowSampleSheetCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.raredisease import (
    RarediseaseSampleSheetCreator,
)
from cg.services.analysis_starter.configurator.implementations.microsalt import (
    MicrosaltConfigurator,
)
from cg.services.analysis_starter.configurator.implementations.nextflow import NextflowConfigurator
from cg.store.store import Store


class ConfiguratorFactory:

    def __init__(self, cg_config: CGConfig):
        self.cg_config = cg_config
        self.housekeeper_api: HousekeeperAPI = cg_config.housekeeper_api
        self.lims_api: LimsAPI = cg_config.lims_api
        self.store: Store = cg_config.status_db

    def get_configurator(self, workflow: Workflow) -> Configurator:
        if workflow in NEXTFLOW_WORKFLOWS:
            return self._get_nextflow_configurator(workflow)
        elif workflow == Workflow.MICROSALT:
            return self._get_microsalt_configurator()
        raise NotImplementedError

    def _get_nextflow_configurator(self, workflow: Workflow) -> NextflowConfigurator:
        config_file_creator = self._get_nextflow_config_file_creator(workflow)
        params_file_creator: ParamsFileCreator = self._get_params_file_creator(workflow)
        sample_sheet_creator: NextflowSampleSheetCreator = self._get_sample_sheet_creator(workflow)
        extension: PipelineExtension = self._get_pipeline_extension(workflow)
        return NextflowConfigurator(
            config_file_creator=config_file_creator,
            housekeeper_api=self.housekeeper_api,
            lims=self.lims_api,
            params_file_creator=params_file_creator,
            pipeline_config=self._get_pipeline_config(workflow),
            sample_sheet_creator=sample_sheet_creator,
            store=self.store,
            pipeline_extension=extension,
        )

    def _get_nextflow_config_file_creator(self, workflow: Workflow) -> NextflowConfigFileCreator:
        pipeline_config: CommonAppConfig = self._get_pipeline_config(workflow)
        return NextflowConfigFileCreator(
            account=pipeline_config.slurm.account,
            platform=pipeline_config.platform,
            resources=pipeline_config.resources,
            store=self.store,
            workflow_config_path=pipeline_config.config,
        )

    def _get_params_file_creator(self, workflow: Workflow) -> ParamsFileCreator:
        if workflow == Workflow.RAREDISEASE:
            pipeline_config: CommonAppConfig = self._get_pipeline_config(workflow)
            return RarediseaseParamsFileCreator(
                lims=self.lims_api, store=self.store, params=pipeline_config.params
            )

    def _get_pipeline_config(self, workflow: Workflow) -> CommonAppConfig:
        return getattr(self.cg_config, workflow)

    def _get_sample_sheet_creator(self, workflow: Workflow) -> NextflowSampleSheetCreator:
        if workflow == Workflow.RAREDISEASE:
            return RarediseaseSampleSheetCreator(
                housekeeper_api=self.cg_config.housekeeper_api,
                lims=self.cg_config.lims_api,
                store=self.store,
            )

    def _get_pipeline_extension(self, workflow: Workflow) -> PipelineExtension:
        if workflow == Workflow.RAREDISEASE:
            gene_panel_creator: GenePanelFileCreator = self._get_gene_panel_file_creator(workflow)
            managed_variants_creator: ManagedVariantsFileCreator = (
                self._get_managed_variants_file_creator(workflow)
            )
            return RarediseaseExtension(
                gene_panel_file_creator=gene_panel_creator,
                managed_variants_file_creator=managed_variants_creator,
            )

    def _get_gene_panel_file_creator(self, workflow: Workflow) -> GenePanelFileCreator:
        return GenePanelFileCreator(scout_api=self._get_scout_api(workflow), store=self.store)

    def _get_managed_variants_file_creator(self, workflow: Workflow) -> ManagedVariantsFileCreator:
        return ManagedVariantsFileCreator(scout_api=self._get_scout_api(workflow), store=self.store)

    def _get_scout_api(self, workflow: Workflow) -> ScoutAPI:
        return (
            self.cg_config.scout_api_38
            if workflow == Workflow.NALLO
            else self.cg_config.scout_api_37
        )

    def _get_microsalt_configurator(self) -> MicrosaltConfigurator:
        return MicrosaltConfigurator(
            config_file_creator=self._get_microsalt_config_file_creator(),
            fastq_handler=MicrosaltFastqHandler(
                housekeeper_api=self.housekeeper_api,
                root_dir=Path(self.cg_config.microsalt.root),
                status_db=self.store,
            ),
            lims_api=self.lims_api,
            microsalt_config=self.cg_config.microsalt,
            store=self.store,
        )

    def _get_microsalt_config_file_creator(self) -> MicrosaltConfigFileCreator:
        return MicrosaltConfigFileCreator(
            lims_api=self.lims_api,
            queries_path=self.cg_config.microsalt.queries_path,
            store=self.store,
        )



from enum import StrEnum

from cg.constants.sequencing import SeqLibraryPrepCategory


class AnalysisStatus:
    CANCELLED: str = "cancelled"
    COMPLETED: str = "completed"
    ERROR: str = "error"
    FAILED: str = "failed"
    PENDING: str = "pending"
    RUNNING: str = "running"
    TIMEOUT: str = "timeout"
    QC: str = "qc"


class AnalysisType(StrEnum):
    OTHER: str = "other"
    TGS: str = SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING
    WES: str = SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING
    WGS: str = SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING
    WTS: str = SeqLibraryPrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING



import datetime as dt
from pathlib import Path

from dateutil.parser import parse as parse_datestr


def parse_str_to_datetime(value: str) -> dt.datetime | None:
    if value:
        return parse_datestr(value)


def parse_str_to_path(value: str) -> Path | None:
    if value:
        return Path(value)



from abc import ABC, abstractmethod
from pathlib import Path


class ParamsFileCreator(ABC):

    @staticmethod
    @abstractmethod
    def get_file_path(case_id: str, case_path: Path) -> Path:
        pass

    @abstractmethod
    def create(self, case_id: str, case_path: Path, sample_sheet_path: Path) -> any:
        pass



from pydantic import BaseModel


class LaunchRequest(BaseModel):
    computeEnvId: str
    configProfiles: list[str]
    configText: str
    paramsText: str
    pipeline: str
    preRunScript: str
    pullLatest: bool = False
    resume: bool = False
    revision: str
    runName: str
    sessionId: str | None = None
    stubRun: bool = False
    workDir: str


class WorkflowLaunchRequest(BaseModel):
    """Structured as the request for the /workflow/launch endpoint."""

    launch: LaunchRequest



"""Module for Balsamic UMI Analysis API."""

import logging

from cg.constants import Workflow
from cg.constants.scout import BALSAMIC_UMI_CASE_TAGS
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


class BalsamicUmiAnalysisAPI(BalsamicAnalysisAPI):
    """Handles communication between BALSAMIC processes
    and the rest of CG infrastructure"""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.BALSAMIC_UMI,
    ):
        super().__init__(config=config, workflow=workflow)

    def get_scout_upload_case_tags(self) -> dict:
        """Return Balsamic UMI Scout upload case tags."""
        return BALSAMIC_UMI_CASE_TAGS



"""Module for the HiSeq2500 sequencing times service."""

from datetime import datetime

from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina.file_parsing.sequencing_times.sequencing_time_service import (
    SequencingTimesService,
)


class Hiseq2500SequencingTimesService(SequencingTimesService):
    """Class to get the sequencing times for HiSeq2500 sequencing runs."""

    @staticmethod
    def get_start_time(run_directory_data: IlluminaRunDirectoryData) -> datetime:
        """Get the sequencer start date and time."""
        return run_directory_data.sequenced_at

    @staticmethod
    def get_end_time(run_directory_data: IlluminaRunDirectoryData) -> datetime:
        """Get the sequencer end date and time."""
        return run_directory_data.sequenced_at



"""Service to clean Illumina run directories."""

import logging
from pathlib import Path

from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.time import TWENTY_ONE_DAYS
from cg.exc import (
    HousekeeperBundleVersionMissingError,
    HousekeeperFileMissingError,
    IlluminaCleanRunError,
)
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.store.models import IlluminaSampleSequencingMetrics, IlluminaSequencingRun
from cg.store.store import Store
from cg.utils.files import remove_directory_and_contents
from cg.utils.time import is_directory_older_than_days_old

LOG = logging.getLogger(__name__)


class IlluminaCleanRunsService:
    """
    Handles the cleaning of Illumina runs in the sequencing-runs and demultiplexed-runs directories.
    Requirements for cleaning:
            Sequencing run is older than 21 days
            Sequencing run is backed up
            Sequencing run is in StatusDB
            Sequencing run has sequencing metrics in StatusDB
            Sequencing run has fastq files in Housekeeper or
                Sequencing run has SPRING files in Housekeeper
                and
                Sequencing run has SPRING metadata in Housekeeper
            Sequencing run has a sample sheet in Housekeeper
    """

    def __init__(
        self,
        status_db: Store,
        housekeeper_api: HousekeeperAPI,
        sequencing_run_path: Path,
        dry_run: bool,
    ):
        self.status_db: Store = status_db
        self.hk_api: HousekeeperAPI = housekeeper_api
        self.sequencing_run_dir_data = IlluminaRunDirectoryData(
            sequencing_run_path=sequencing_run_path
        )
        self.dry_run: bool = dry_run
        LOG.info(f"Trying to delete {sequencing_run_path}")

    def delete_run_directory(self) -> None:
        """
        Delete the sequencing runs and demultiplexed runs directory if it fulfills all requirements.
        Raises:
            IlluminaCleanRunError when any exception is caught
        """
        try:
            self.set_sample_sheet_path_from_housekeeper()
            if self.can_run_directory_be_deleted():
                if self.dry_run:
                    LOG.debug(f"Dry run: Would have removed: {self.sequencing_run_dir_data.path}")
                    return
                remove_directory_and_contents(self.sequencing_run_dir_data.path)
        except Exception as error:
            raise IlluminaCleanRunError(
                f"Sequencing run with path {self.sequencing_run_dir_data.path} not removed: {repr(error)}"
            )

    def set_sample_sheet_path_from_housekeeper(self):
        """
        Set the sample sheet for a sequencing run.
        Raises:
            HousekeeperFileMissingError when the sample sheet is missing in Housekeeper
        """
        sample_sheet_path: Path = self.hk_api.get_sample_sheet_path(self.sequencing_run_dir_data.id)
        self.sequencing_run_dir_data.set_sample_sheet_path_hk(sample_sheet_path)

    def can_run_directory_be_deleted(self) -> bool:
        """Determine whether a sequencing run or demultiplexed run can be deleted."""
        return all(
            [
                self.is_directory_older_than_21_days(),
                self.is_sequencing_run_in_statusdb(),
                self.is_sequencing_run_backed_up(),
                self.has_sequencing_metrics_in_statusdb(),
                self.has_sample_fastq_or_spring_files_in_housekeeper(),
                self.has_sample_sheet_in_housekeeper(),
            ]
        )

    def is_directory_older_than_21_days(self) -> bool:
        """Check if a given directory is older than 21 days."""
        return is_directory_older_than_days_old(
            directory_path=self.sequencing_run_dir_data.path,
            days_old=TWENTY_ONE_DAYS,
        )

    def is_sequencing_run_in_statusdb(self) -> bool:
        """Check if sequencing run is in StatusDB."""
        return bool(self.get_sequencing_run_from_status_db())

    def is_sequencing_run_backed_up(self) -> bool:
        """Check if sequencing run is backed up on PDC."""
        return self.get_sequencing_run_from_status_db().has_backup

    def has_sequencing_metrics_in_statusdb(self) -> bool:
        """Check if a sequencing run has entries in the IlluminaSampleSequencingMetrics table."""
        return bool(self.get_sequencing_metrics_for_sequencing_run())

    def has_sample_sheet_in_housekeeper(self) -> bool:
        """Check if the sequencing run has a sample sheet in Housekeeper."""
        return bool(self.sequencing_run_dir_data.get_sample_sheet_path_hk())

    def has_fastq_files_for_samples_in_housekeeper(self) -> bool:
        """Check if all samples on the sequencing run have fastq files in Housekeeper."""
        return bool(self.get_files_for_samples_on_flow_cell_with_tag(tag=SequencingFileTag.FASTQ))

    def has_spring_files_for_samples_in_housekeeper(self) -> bool:
        """Check if all samples on the sequencing run have SPRING files in Housekeeper."""
        return bool(self.get_files_for_samples_on_flow_cell_with_tag(tag=SequencingFileTag.SPRING))

    def has_spring_meta_data_files_for_samples_in_housekeeper(self) -> bool:
        """Check if all samples on the sequencing run have SPRING metadata files in Housekeeper."""
        return bool(
            self.get_files_for_samples_on_flow_cell_with_tag(tag=SequencingFileTag.SPRING_METADATA)
        )

    def has_sample_fastq_or_spring_files_in_housekeeper(self) -> bool:
        """
        Check if a flow cell has fastq or spring files in Housekeeper.
        Raises:
            HousekeeperFileMissingError
        """
        if not self.has_fastq_files_for_samples_in_housekeeper() and not (
            self.has_spring_files_for_samples_in_housekeeper()
            and self.has_spring_meta_data_files_for_samples_in_housekeeper()
        ):
            raise HousekeeperFileMissingError(
                f"sequencing run {self.sequencing_run_dir_data.id} is missing fastq and spring files for some samples."
            )
        return True

    def get_sequencing_run_from_status_db(self) -> IlluminaSequencingRun | None:
        """
        Get the sequencing run entry from StatusDB.
        Raises:
            ValueError if the sequencing run is not found in StatusDB.
        """
        sequencing_run: IlluminaSequencingRun = (
            self.status_db.get_illumina_sequencing_run_by_device_internal_id(
                self.sequencing_run_dir_data.id
            )
        )
        return sequencing_run

    def get_sequencing_metrics_for_sequencing_run(
        self,
    ) -> list[IlluminaSampleSequencingMetrics] | None:
        """
        Get the sample sequencing metrics entries for a sequencing run.
        Raises:
              Value error if no IlluminaSampleSequencingMetrics are found in StatusDB.
        """
        sequencing_run: IlluminaSequencingRun = self.get_sequencing_run_from_status_db()
        metrics: list[IlluminaSampleSequencingMetrics] = sequencing_run.sample_metrics
        if not metrics:
            raise ValueError(
                f"No sample sequencing metrics found for {self.sequencing_run_dir_data.id} in StatusDB."
            )
        return metrics

    def get_files_for_samples_on_flow_cell_with_tag(self, tag: str) -> list[File] | None:
        """Return the files with the specified tag for all samples on a sequencing run."""
        sequencing_run: IlluminaSequencingRun = self.get_sequencing_run_from_status_db()
        bundle_names: list[str] = [
            metrics.sample.internal_id for metrics in sequencing_run.sample_metrics
        ]
        files: list[File] = []
        for bundle_name in bundle_names:
            try:
                files.extend(
                    self.hk_api.get_files_from_latest_version(
                        bundle_name=bundle_name, tags=[tag, self.sequencing_run_dir_data.id]
                    )
                )
            except HousekeeperBundleVersionMissingError:
                LOG.warning(
                    f"No files with tag {tag} found on flow cell {self.sequencing_run_dir_data.id}"
                )
                LOG.warning(f"Bundle: {bundle_name} not found in Housekeeper.")
                return None
        return files



"""Utility functions for the post-process command."""

import logging

from pydantic import BaseModel, ConfigDict

from cg.exc import CgError
from cg.models.cg_config import CGConfig
from cg.services.run_devices.abstract_classes import PostProcessingService
from cg.services.run_devices.pacbio.post_processing_service import PacBioPostProcessingService
from cg.services.run_devices.run_names.service import RunNamesService
from cg.utils.mapping import get_item_by_pattern_in_source

LOG = logging.getLogger(__name__)

PATTERN_TO_DEVICE_MAP: dict[str, str] = {
    r"^r\d+_\d+_\d+/(1|2)_[^/]+$": "pacbio",
}


class UnprocessedRunInfo(BaseModel):
    name: str
    post_processing_service: PostProcessingService
    instrument: str
    model_config = ConfigDict(arbitrary_types_allowed=True)


def get_post_processing_service_from_run_name(
    context: CGConfig, run_name: str
) -> PacBioPostProcessingService:
    """Get the correct post-processing service based on the run name."""
    try:
        device: str = get_item_by_pattern_in_source(
            source=run_name, pattern_map=PATTERN_TO_DEVICE_MAP
        )
    except CgError as error:
        raise NameError(
            f"Run name {run_name} does not match with any known sequencing run name pattern"
        ) from error
    return getattr(context.post_processing_services, device)


def get_unprocessed_runs_info(context: CGConfig, instrument: str) -> list[UnprocessedRunInfo]:
    """Return a list of unprocessed runs for a given instrument or for all instruments."""
    runs: list[UnprocessedRunInfo] = []
    instruments_to_check: list[str] = _instruments_to_check(instrument)
    for instrument_name in instruments_to_check:
        run_names_service: RunNamesService = getattr(context.run_names_services, instrument_name)
        runs.extend(
            _get_unprocessed_runs_from_run_names(
                run_names=run_names_service.get_run_names(),
                post_processing_service=getattr(context.post_processing_services, instrument_name),
                instrument_name=instrument_name,
            )
        )
    return runs


def _instruments_to_check(instrument: str) -> list[str]:
    """Return a list of instruments to check for unprocessed runs."""
    possible_instruments: list[str] = ["pacbio"]  # Add more instruments here
    return [instrument] if instrument != "all" else possible_instruments


def _get_unprocessed_runs_from_run_names(
    run_names: list[str], post_processing_service: PostProcessingService, instrument_name
) -> list[UnprocessedRunInfo]:
    LOG.debug(f"Adding {instrument_name} run names to the post-processing list")
    runs: list[UnprocessedRunInfo] = []
    for name in run_names:
        if post_processing_service.is_run_processed(name):
            LOG.debug(f"Run {name} has already been post-processed. Skipping")
            continue
        if post_processing_service.can_post_processing_start(name):
            runs.append(
                UnprocessedRunInfo(
                    name=name,
                    post_processing_service=post_processing_service,
                    instrument=instrument_name,
                )
            )
    return runs



from pydantic import BaseModel


class OrderError(BaseModel):
    field: str
    message: str


class UserNotAssociatedWithCustomerError(OrderError):
    field: str = "customer"
    message: str = "User does not belong to customer"


class CustomerCannotSkipReceptionControlError(OrderError):
    field: str = "skip_reception_control"
    message: str = "Customer cannot skip reception control"


class CustomerDoesNotExistError(OrderError):
    field: str = "customer"
    message: str = "Customer does not exist"


class OrderNameRequiredError(OrderError):
    field: str = "name"
    message: str = "Order name is required"



from cg.services.orders.validation.models.case_aliases import CaseContainingRelatives
from cg.services.orders.validation.models.existing_sample import ExistingSample
from cg.services.orders.validation.order_types.mip_dna.models.sample import MIPDNASample
from cg.services.orders.validation.order_types.tomte.models.sample import TomteSample
from cg.store.store import Store

SampleWithParents = TomteSample | MIPDNASample | ExistingSample


class Node:
    """
    This class is used to represent the samples in the family tree graph. The variables 'mother' and
    'father' refer to other nodes in the family tree, and can be thought of as an edge in the graph.
    Because the 'mother' and 'father' are tracked using the sample's _name_ in the order, and
    because said name is not set in the ExistingSample model, we require the sample name as a
    separate input.
    """

    def __init__(
        self,
        sample: SampleWithParents,
        case_index: int,
        sample_index: int,
        sample_name: str,
    ):
        self.sample: SampleWithParents = sample
        self.sample_name: str = sample_name
        self.sample_index: int = sample_index
        self.case_index: int = case_index
        self.father: Node | None = None
        self.mother: Node | None = None
        self.visited = False
        self.in_current_path = False


class FamilyTree:
    """
    This class is a directed graph representing a family tree from a submitted order with specified
    mothers and fathers. Each node represents a sample, and each node has a property 'mother' and
    a property 'father' referring to other nodes in the graph. These may be thought of as the
    graph's edges.
    """

    def __init__(self, case: CaseContainingRelatives, case_index: int, store: Store):
        self.graph: dict[str, Node] = {}
        self.case: CaseContainingRelatives = case
        self.case_index: int = case_index
        self.store = store
        self._add_nodes()
        self._add_edges()

    def _add_nodes(self) -> None:
        """Add a node to the graph for each sample in the graph. For existing samples, the name
        is fetched from StatusDB."""
        for sample_index, sample in self.case.enumerated_samples:
            if sample.is_new:
                sample_name = sample.name
            else:
                sample_name = self.store.get_sample_by_internal_id(sample.internal_id).name
            node = Node(
                sample=sample,
                sample_index=sample_index,
                case_index=self.case_index,
                sample_name=sample_name,
            )
            self.graph[sample_name] = node

    def _add_edges(self) -> None:
        """Add edges to the graph by populating each node's 'mother' and 'father' property."""
        for node in self.graph.values():
            sample: SampleWithParents = node.sample
            if sample.mother:
                node.mother = self.graph.get(sample.mother)
            if sample.father:
                node.father = self.graph.get(sample.father)

    @property
    def nodes(self) -> list[Node]:
        return list(self.graph.values())



""" Module for retrieving flow cells from backup."""

import logging
import subprocess
from pathlib import Path

from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.exc import ChecksumFailedError
from cg.meta.encryption.encryption import SpringEncryptionAPI
from cg.models.compression_data import CompressionData
from cg.services.pdc_service.pdc_service import PdcService

LOG = logging.getLogger(__name__)


class SpringBackupAPI:
    """Class for handling PDC backup and retrieval of spring compressed fastq files."""

    def __init__(
        self,
        encryption_api: SpringEncryptionAPI,
        pdc_service: PdcService,
        hk_api: HousekeeperAPI,
        dry_run: bool = False,
    ):
        self.encryption_api: SpringEncryptionAPI = encryption_api
        self.hk_api: HousekeeperAPI = hk_api
        self.pdc: PdcService = pdc_service
        self.dry_run = dry_run

    def encrypt_and_archive_spring_file(self, spring_file_path: Path) -> None:
        """Encrypts and archives a spring file and its decryption key."""
        LOG.debug(f"*** START BACKUP PROCESS OF SPRING FILE {spring_file_path} ***")
        if self.is_compression_ongoing(spring_file_path):
            LOG.info(
                f"Spring (de)compression ongoing, skipping archiving for spring file {spring_file_path}",
            )
            return
        self.encryption_api.cleanup(spring_file_path)
        if self.is_spring_file_archived(spring_file_path):
            LOG.info(
                "Spring file already archived! Spring file will be removed from disk, continuing "
                "to next spring file "
            )
            self.remove_archived_spring_file(spring_file_path)
            return
        try:
            self.encryption_api.spring_symmetric_encryption(spring_file_path)
            self.encryption_api.key_asymmetric_encryption(spring_file_path)
            self.encryption_api.compare_spring_file_checksums(spring_file_path)
            self.pdc.archive_file_to_pdc(
                file_path=str(self.encryption_api.encrypted_spring_file_path(spring_file_path))
            )
            self.pdc.archive_file_to_pdc(
                file_path=str(self.encryption_api.encrypted_key_path(spring_file_path))
            )
            self.mark_file_as_archived(spring_file_path)
            self.encryption_api.cleanup(spring_file_path)
            self.remove_archived_spring_file(spring_file_path)
            LOG.debug("*** ARCHIVING PROCESS COMPLETED SUCCESSFULLY ***")
        except subprocess.CalledProcessError as error:
            LOG.error(f"Encryption failed: {error.stderr}")
            LOG.debug("*** COMMAND PROCESS FAILED! ***")
            self.encryption_api.cleanup(spring_file_path)
        except ChecksumFailedError as error:
            LOG.error(error)
            self.encryption_api.cleanup(spring_file_path)
            LOG.debug("*** CHECKSUM PROCESS FAILED! ***")

    def retrieve_and_decrypt_spring_file(self, spring_file_path: Path) -> None:
        """Retrieves and decrypts a spring file and its decryption key."""
        LOG.info(f"*** START RETRIEVAL PROCESS OF SPRING FILE {spring_file_path} ***")
        try:
            self.pdc.retrieve_file_from_pdc(
                file_path=str(self.encryption_api.encrypted_spring_file_path(spring_file_path)),
            )
            self.pdc.retrieve_file_from_pdc(
                file_path=str(self.encryption_api.encrypted_key_path(spring_file_path)),
            )
            self.encryption_api.key_asymmetric_decryption(spring_file_path)
            self.encryption_api.spring_symmetric_decryption(
                spring_file_path, output_file=spring_file_path
            )
        except subprocess.CalledProcessError as error:
            LOG.error(f"Decryption failed: {error.stderr}")
            LOG.debug("*** RETRIEVAL PROCESS FAILED! ***")
        self.encryption_api.cleanup(spring_file_path)
        LOG.debug("*** RETRIEVAL PROCESS COMPLETED SUCCESSFULLY ***")

    def mark_file_as_archived(self, spring_file_path: Path) -> None:
        """Set the field 'to_archive' of the file in Housekeeper to mark that it has been
        archived to PDC."""
        if self.dry_run:
            LOG.info(f"Dry run, no changes made to {spring_file_path}")
            return
        hk_spring_file: File = self.hk_api.files(path=str(spring_file_path)).first()
        if hk_spring_file:
            LOG.info(f"Setting {spring_file_path} to archived in Housekeeper")
            self.hk_api.set_to_archive(file=hk_spring_file, value=True)
        else:
            LOG.warning(f"Could not find {spring_file_path} on disk")

    def remove_archived_spring_file(self, spring_file_path: Path) -> None:
        """Removes all files related to spring PDC archiving."""
        if not self.dry_run:
            LOG.info(f"Removing spring file {spring_file_path} from disk")
            spring_file_path.unlink()

    def is_to_be_retrieved_and_decrypted(self, spring_file_path: Path) -> bool:
        """Determines if a spring file is archived on PDC and needs to be retrieved and decrypted."""
        spring_file: File = self.hk_api.files(path=str(spring_file_path)).first()
        if spring_file and not spring_file_path.exists():
            return spring_file.to_archive
        return False

    def is_spring_file_archived(self, spring_file_path: Path) -> bool:
        """Checks if a spring file is marked as archived in Housekeeper."""
        spring_file: File = self.hk_api.files(path=str(spring_file_path)).first()
        if spring_file:
            return spring_file.to_archive
        return False

    @staticmethod
    def is_compression_ongoing(spring_file_path: Path) -> bool:
        """Determines if (de)compression of the spring file ongoing."""
        return CompressionData(spring_file_path).pending_exists()



import logging
from subprocess import CalledProcessError

from cg.constants import Workflow
from cg.exc import AnalysisNotReadyError, AnalysisRunningError
from cg.services.analysis_starter.configurator.abstract_model import CaseConfig
from cg.services.analysis_starter.configurator.configurator import Configurator
from cg.services.analysis_starter.input_fetcher.input_fetcher import InputFetcher
from cg.services.analysis_starter.submitters.submitter import Submitter
from cg.services.analysis_starter.tracker.tracker import Tracker
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class AnalysisStarter:
    def __init__(
        self,
        configurator: Configurator,
        input_fetcher: InputFetcher,
        store: Store,
        submitter: Submitter,
        tracker: Tracker,
        workflow: Workflow,
    ):
        self.configurator = configurator
        self.input_fetcher = input_fetcher
        self.store = store
        self.submitter = submitter
        self.tracker = tracker
        self.workflow = workflow

    def start_available(self, limit: int | None = None) -> bool:
        """Starts available cases. Returns True if all ready cases started without an error."""
        succeeded = True
        cases: list[Case] = self.store.get_cases_to_analyze(workflow=self.workflow, limit=limit)
        for case in cases:
            try:
                self.start(case.internal_id)
            except AnalysisNotReadyError as error:
                LOG.error(error)
            except AnalysisRunningError as error:
                LOG.error(error)
                succeeded = False
            except Exception as error:
                LOG.error(f"Unspecified error occurred: {error}")
                succeeded = False
        return succeeded

    def start(self, case_id: str, **flags) -> None:
        """Fetches raw data, generates configuration files and runs the specified case."""
        self.tracker.ensure_analysis_not_ongoing(case_id)
        self.input_fetcher.ensure_files_are_ready(case_id)
        parameters: CaseConfig = self.configurator.configure(case_id=case_id, **flags)
        self._run_and_track(case_id=case_id, parameters=parameters)

    def run(self, case_id: str, **flags) -> None:
        """Run a case using an assumed existing configuration."""
        self.tracker.ensure_analysis_not_ongoing(case_id)
        parameters: CaseConfig = self.configurator.get_config(case_id=case_id, **flags)
        self._run_and_track(case_id=case_id, parameters=parameters)

    def _run_and_track(self, case_id: str, parameters: CaseConfig):
        self.tracker.set_case_as_running(case_id)
        try:
            tower_workflow_id: str | None = self.submitter.submit(parameters)
            self.tracker.track(case_config=parameters, tower_workflow_id=tower_workflow_id)
        except CalledProcessError as exception:
            self.tracker.set_case_as_not_running(case_id)
            raise exception
